import pytest

import geomstats.backend as gs
from geomstats.geometry.special_orthogonal import (
    SpecialOrthogonal,
    _SpecialOrthogonal2Vectors,
    _SpecialOrthogonal3Vectors,
)
from geomstats.test.random import (
    LieGroupVectorRandomDataGenerator,
    get_random_quaternion,
)
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.base import LevelSetTestCase
from geomstats.test_cases.geometry.lie_group import (
    LieGroupTestCase,
    MatrixLieGroupTestCase,
)
from geomstats.test_cases.geometry.mixins import ProjectionTestCaseMixins

# TODO: improve random generation


class _SpecialOrthogonalTestCaseMixins:
    def _get_random_rotation_vector(self, n_points=1):
        if self.space.n == 2:
            return _SpecialOrthogonal2Vectors(equip=False).random_point(n_points)

        if self.space.n == 3:
            return _SpecialOrthogonal3Vectors(equip=False).random_point(n_points)

        raise NotImplementedError(
            f"Unable to create random orthogonal vector for `n={self.space.n}`"
        )

    def _get_random_rotation_matrix(self, n_points=1):
        return SpecialOrthogonal(n=self.space.n, equip=False).random_point(n_points)

    def test_rotation_vector_from_matrix(self, rot_mat, expected, atol):
        rot_vec = self.space.rotation_vector_from_matrix(rot_mat)
        self.assertAllClose(rot_vec, expected, atol=atol)

    @pytest.mark.vec
    def test_rotation_vector_from_matrix_vec(self, n_reps, atol):
        rot_mat = self._get_random_rotation_matrix()

        expected = self.space.rotation_vector_from_matrix(rot_mat)

        vec_data = generate_vectorization_data(
            data=[dict(rot_mat=rot_mat, expected=expected, atol=atol)],
            arg_names=["rot_mat"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_matrix_from_rotation_vector(self, rot_vec, expected, atol):
        rot_mat = self.space.matrix_from_rotation_vector(rot_vec)
        self.assertAllClose(rot_mat, expected, atol=atol)

    @pytest.mark.vec
    def test_matrix_from_rotation_vector_vec(self, n_reps, atol):
        rot_vec = self._get_random_rotation_vector()
        expected = self.space.matrix_from_rotation_vector(rot_vec)

        vec_data = generate_vectorization_data(
            data=[dict(rot_vec=rot_vec, expected=expected, atol=atol)],
            arg_names=["rot_vec"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_rotation_vector_from_matrix_after_matrix_from_rotation_vector(
        self, n_points, atol
    ):
        vec = self._get_random_rotation_vector(n_points)
        mat = self.space.matrix_from_rotation_vector(vec)
        vec_ = self.space.rotation_vector_from_matrix(mat)
        self.assertAllClose(vec, vec_, atol=atol)

    @pytest.mark.random
    def test_matrix_from_rotation_vector_after_rotation_vector_from_matrix(
        self, n_points, atol
    ):
        mat = self._get_random_rotation_matrix(n_points)
        vec = self.space.rotation_vector_from_matrix(mat)
        mat_ = self.space.matrix_from_rotation_vector(vec)
        self.assertAllClose(mat, mat_, atol=atol)


class SpecialOrthogonalMatricesTestCase(
    _SpecialOrthogonalTestCaseMixins, MatrixLieGroupTestCase, LevelSetTestCase
):
    def test_are_antipodals(self, rotation_mat1, rotation_mat2, expected, atol):
        res = self.space.are_antipodals(rotation_mat1, rotation_mat2)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_are_antipodals_vec(self, n_reps, atol):
        rotation_mat1 = self._get_random_rotation_matrix()
        rotation_mat2 = self._get_random_rotation_matrix()

        expected = self.space.are_antipodals(rotation_mat1, rotation_mat2)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    rotation_mat1=rotation_mat1,
                    rotation_mat2=rotation_mat2,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["rotation_mat1", "rotation_mat2"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)


class SpecialOrthogonalVectorsTestCase(
    ProjectionTestCaseMixins, _SpecialOrthogonalTestCaseMixins, LieGroupTestCase
):
    # TODO: add test on projection matrix belongs?
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = LieGroupVectorRandomDataGenerator(self.space)

    def _get_random_rotation_vector(self, n_points=1):
        return self.space.random_point(n_points)

    @pytest.mark.vec
    def test_projection_vec(self, n_reps, atol):
        # TODO: review class code design

        point = gs.random.normal(size=(self.space.n, self.space.n))
        proj_point = self.space.projection(point)

        vec_data = generate_vectorization_data(
            data=[dict(point=point, expected=proj_point, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)


class SpecialOrthogonal3VectorsTestCase(SpecialOrthogonalVectorsTestCase):
    def _assert_quaternion(self, quaternion_, quaternion, atol):
        self.assertAllClose(gs.abs(quaternion_), gs.abs(quaternion), atol=atol)

    def _assert_tait_bryan_angles(self, angles_, angles, extrinsic, zyx, atol):
        try:
            self.assertAllClose(angles_, angles, atol=atol)
        except AssertionError:
            mat_ = self.space.matrix_from_tait_bryan_angles(
                angles_, extrinsic=extrinsic, zyx=zyx
            )
            mat = self.space.matrix_from_tait_bryan_angles(
                angles, extrinsic=extrinsic, zyx=zyx
            )
            self.assertAllClose(mat_, mat, atol=atol)

    def _get_random_angles(self, n_points=1):
        size = (n_points, 3) if n_points > 1 else 3
        return gs.random.uniform(low=-gs.pi, high=gs.pi, size=size)

    def test_group_exp_after_log_with_angles_close_to_pi(self, point, base_point):
        result = self.space.exp(self.space.log(point, base_point), base_point)
        expected = self.space.regularize(point)
        inv_expected = -expected

        self.assertTrue(
            gs.allclose(result, expected, atol=5e-3)
            or gs.allclose(result, inv_expected, atol=5e-3)
        )

    def test_group_log_after_exp_with_angles_close_to_pi(self, tangent_vec, base_point):
        result = self.space.log(self.space.exp(tangent_vec, base_point), base_point)
        reg_tangent_vec = self.space.regularize_tangent_vec(
            tangent_vec=tangent_vec, base_point=base_point
        )
        expected = reg_tangent_vec
        inv_expected = -expected
        self.assertTrue(
            gs.allclose(result, expected, atol=5e-3)
            or gs.allclose(result, inv_expected, atol=5e-3)
        )

    def test_rotation_vector_and_rotation_matrix_with_angles_close_to_pi(self, point):
        mat = self.space.matrix_from_rotation_vector(point)
        result = self.space.rotation_vector_from_matrix(mat)
        expected1 = self.space.regularize(point)
        expected2 = -1 * expected1
        expected = gs.allclose(result, expected1) or gs.allclose(result, expected2)
        self.assertTrue(expected)

    def test_quaternion_from_matrix(self, rot_mat, expected, atol):
        res = self.space.quaternion_from_matrix(rot_mat)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_quaternion_from_matrix_vec(self, n_reps, atol):
        rot_mat = self._get_random_rotation_matrix()
        expected = self.space.quaternion_from_matrix(rot_mat)

        vec_data = generate_vectorization_data(
            data=[dict(rot_mat=rot_mat, expected=expected, atol=atol)],
            arg_names=["rot_mat"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_matrix_from_quaternion(self, quaternion, expected, atol):
        res = self.space.matrix_from_quaternion(quaternion)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_matrix_from_quaternion_vec(self, n_reps, atol):
        quaternion = get_random_quaternion()
        expected = self.space.matrix_from_quaternion(quaternion)

        vec_data = generate_vectorization_data(
            data=[dict(quaternion=quaternion, expected=expected, atol=atol)],
            arg_names=["quaternion"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_quaternion_from_matrix_after_matrix_from_quaternion(self, n_points, atol):
        quaternion = get_random_quaternion(n_points)
        mat = self.space.matrix_from_quaternion(quaternion)
        quaternion_ = self.space.quaternion_from_matrix(mat)

        self._assert_quaternion(quaternion_, quaternion, atol)

    @pytest.mark.random
    def test_matrix_from_quaternion_after_quaternion_from_matrix(self, n_points, atol):
        mat = self._get_random_rotation_matrix(n_points)
        quaternion = self.space.quaternion_from_matrix(mat)
        mat_ = self.space.matrix_from_quaternion(quaternion)
        self.assertAllClose(mat_, mat, atol=atol)

    def test_quaternion_and_matrix_with_angles_close_to_pi(self, point):
        mat = self.space.matrix_from_rotation_vector(point)
        quat = self.space.quaternion_from_matrix(mat)
        result = self.space.matrix_from_quaternion(quat)
        expected1 = mat
        expected2 = gs.linalg.inv(mat)
        expected = gs.allclose(result, expected1) or gs.allclose(result, expected2)
        self.assertTrue(expected)

    def test_quaternion_from_rotation_vector(self, rot_vec, expected, atol):
        res = self.space.quaternion_from_rotation_vector(rot_vec)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_quaternion_from_rotation_vector_vec(self, n_reps, atol):
        rot_vec = self._get_random_rotation_vector()
        expected = self.space.quaternion_from_rotation_vector(rot_vec)

        vec_data = generate_vectorization_data(
            data=[dict(rot_vec=rot_vec, expected=expected, atol=atol)],
            arg_names=["rot_vec"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_rotation_vector_from_quaternion(self, quaternion, expected, atol):
        res = self.space.rotation_vector_from_quaternion(quaternion)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_rotation_vector_from_quaternion_vec(self, n_reps, atol):
        quaternion = get_random_quaternion()
        expected = self.space.rotation_vector_from_quaternion(quaternion)

        vec_data = generate_vectorization_data(
            data=[dict(quaternion=quaternion, expected=expected, atol=atol)],
            arg_names=["quaternion"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_quaternion_from_rotation_vector_after_rotation_vector_from_quaternion(
        self, n_points, atol
    ):
        quaternion = get_random_quaternion(n_points)
        rot_vec = self.space.rotation_vector_from_quaternion(quaternion)
        quaternion_ = self.space.quaternion_from_rotation_vector(rot_vec)

        self._assert_quaternion(quaternion_, quaternion, atol)

    @pytest.mark.random
    def test_rotation_vector_from_quaternion_after_quaternion_from_rotation_vector(
        self, n_points, atol
    ):
        rot_vec = self._get_random_rotation_vector(n_points)
        quaternion = self.space.quaternion_from_rotation_vector(rot_vec)
        rot_vec_ = self.space.rotation_vector_from_quaternion(quaternion)
        self.assertAllClose(rot_vec_, rot_vec, atol=atol)

    def test_quaternion_and_rotation_vector_with_angles_close_to_pi(self, point):
        quaternion = self.space.quaternion_from_rotation_vector(point)
        result = self.space.rotation_vector_from_quaternion(quaternion)
        expected1 = self.space.regularize(point)
        expected2 = -1 * expected1
        expected = gs.allclose(result, expected1) or gs.allclose(result, expected2)
        self.assertTrue(expected)

    def test_matrix_from_tait_bryan_angles(
        self, tait_bryan_angles, extrinsic, zyx, expected, atol
    ):
        res = self.space.matrix_from_tait_bryan_angles(
            tait_bryan_angles, extrinsic=extrinsic, zyx=zyx
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_matrix_from_tait_bryan_angles_vec(self, n_reps, extrinsic, zyx, atol):
        angles = self._get_random_angles()
        expected = self.space.matrix_from_tait_bryan_angles(
            angles, extrinsic=extrinsic, zyx=zyx
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tait_bryan_angles=angles,
                    extrinsic=extrinsic,
                    zyx=zyx,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tait_bryan_angles"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_tait_bryan_angles_from_matrix(
        self, rot_mat, extrinsic, zyx, expected, atol
    ):
        res = self.space.tait_bryan_angles_from_matrix(
            rot_mat, extrinsic=extrinsic, zyx=zyx
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_tait_bryan_angles_from_matrix_vec(self, n_reps, extrinsic, zyx, atol):
        rot_mat = self._get_random_rotation_matrix()
        expected = self.space.tait_bryan_angles_from_matrix(
            rot_mat, extrinsic=extrinsic, zyx=zyx
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    rot_mat=rot_mat,
                    extrinsic=extrinsic,
                    zyx=zyx,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["rot_mat"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_tait_bryan_angles_from_matrix_after_matrix_from_tait_bryan_angles(
        self, n_points, extrinsic, zyx, atol
    ):
        angles = self._get_random_angles(n_points)
        mat = self.space.matrix_from_tait_bryan_angles(
            angles, extrinsic=extrinsic, zyx=zyx
        )
        angles_ = self.space.tait_bryan_angles_from_matrix(
            mat, extrinsic=extrinsic, zyx=zyx
        )
        # self.assertAllClose(angles_, angles, atol=atol)
        self._assert_tait_bryan_angles(
            angles_, angles, extrinsic=extrinsic, zyx=zyx, atol=atol
        )

    @pytest.mark.random
    def test_matrix_from_tait_bryan_angles_after_tait_bryan_angles_from_matrix(
        self, n_points, extrinsic, zyx, atol
    ):
        mat = self._get_random_rotation_matrix(n_points)
        angles = self.space.tait_bryan_angles_from_matrix(
            mat, extrinsic=extrinsic, zyx=zyx
        )
        mat_ = self.space.matrix_from_tait_bryan_angles(
            angles, extrinsic=extrinsic, zyx=zyx
        )

        self.assertAllClose(mat_, mat, atol=atol)

    def test_quaternion_from_tait_bryan_angles(
        self, tait_bryan_angles, extrinsic, zyx, expected, atol
    ):
        res = self.space.quaternion_from_tait_bryan_angles(
            tait_bryan_angles, extrinsic=extrinsic, zyx=zyx
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_quaternion_from_tait_bryan_angles_vec(self, n_reps, extrinsic, zyx, atol):
        angles = self._get_random_angles()
        expected = self.space.quaternion_from_tait_bryan_angles(
            angles, extrinsic=extrinsic, zyx=zyx
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tait_bryan_angles=angles,
                    extrinsic=extrinsic,
                    zyx=zyx,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tait_bryan_angles"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_tait_bryan_angles_from_quaternion(
        self, quaternion, extrinsic, zyx, expected, atol
    ):
        res = self.space.tait_bryan_angles_from_quaternion(
            quaternion, extrinsic=extrinsic, zyx=zyx
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_tait_bryan_angles_from_quaternion_vec(self, n_reps, extrinsic, zyx, atol):
        quaternion = get_random_quaternion()
        expected = self.space.tait_bryan_angles_from_quaternion(
            quaternion, extrinsic=extrinsic, zyx=zyx
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    quaternion=quaternion,
                    extrinsic=extrinsic,
                    zyx=zyx,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["quaternion"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_quaternion_from_tait_bryan_angles_after_tait_bryan_angles_from_quaternion(
        self, n_points, extrinsic, zyx, atol
    ):
        quaternion = get_random_quaternion(n_points)
        angles = self.space.tait_bryan_angles_from_quaternion(
            quaternion, extrinsic=extrinsic, zyx=zyx
        )
        quaternion_ = self.space.quaternion_from_tait_bryan_angles(
            angles, extrinsic=extrinsic, zyx=zyx
        )
        self._assert_quaternion(quaternion_, quaternion, atol=atol)

    @pytest.mark.random
    def test_tait_bryan_angles_from_quaternion_after_quaternion_from_tait_bryan_angles(
        self, n_points, extrinsic, zyx, atol
    ):
        angles = self._get_random_angles(n_points)
        quaternion = self.space.quaternion_from_tait_bryan_angles(
            angles, extrinsic=extrinsic, zyx=zyx
        )
        angles_ = self.space.tait_bryan_angles_from_quaternion(
            quaternion, extrinsic=extrinsic, zyx=zyx
        )

        self._assert_tait_bryan_angles(
            angles_, angles, extrinsic=extrinsic, zyx=zyx, atol=atol
        )

    def test_rotation_vector_from_tait_bryan_angles(
        self, tait_bryan_angles, extrinsic, zyx, expected, atol
    ):
        res = self.space.rotation_vector_from_tait_bryan_angles(
            tait_bryan_angles, extrinsic=extrinsic, zyx=zyx
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_rotation_vector_from_tait_bryan_angles_vec(
        self, n_reps, extrinsic, zyx, atol
    ):
        angles = self._get_random_angles()
        expected = self.space.rotation_vector_from_tait_bryan_angles(
            angles, extrinsic=extrinsic, zyx=zyx
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tait_bryan_angles=angles,
                    extrinsic=extrinsic,
                    zyx=zyx,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tait_bryan_angles"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_tait_bryan_angles_from_rotation_vector(
        self, rot_vec, extrinsic, zyx, expected, atol
    ):
        res = self.space.tait_bryan_angles_from_rotation_vector(
            rot_vec, extrinsic=extrinsic, zyx=zyx
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_tait_bryan_angles_from_rotation_vector_vec(
        self, n_reps, extrinsic, zyx, atol
    ):
        rot_vec = self._get_random_rotation_vector()
        expected = self.space.tait_bryan_angles_from_rotation_vector(
            rot_vec, extrinsic=extrinsic, zyx=zyx
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    rot_vec=rot_vec,
                    extrinsic=extrinsic,
                    zyx=zyx,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["rot_vec"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_tait_bryan_angles_from_rotation_vector_after_rotation_vector_from_tait_bryan_angles(
        self, n_points, extrinsic, zyx, atol
    ):
        angles = self._get_random_angles(n_points)
        rot_vec = self.space.rotation_vector_from_tait_bryan_angles(
            angles, extrinsic=extrinsic, zyx=zyx
        )
        angles_ = self.space.tait_bryan_angles_from_rotation_vector(
            rot_vec, extrinsic=extrinsic, zyx=zyx
        )

        self._assert_tait_bryan_angles(
            angles_, angles, extrinsic=extrinsic, zyx=zyx, atol=atol
        )

    @pytest.mark.random
    def test_rotation_vector_from_tait_bryan_angles_after_tait_bryan_angles_from_rotation_vector(
        self, n_points, extrinsic, zyx, atol
    ):
        rot_vec = self._get_random_rotation_vector(n_points)
        angles = self.space.tait_bryan_angles_from_rotation_vector(
            rot_vec, extrinsic=extrinsic, zyx=zyx
        )
        rot_vec_ = self.space.rotation_vector_from_tait_bryan_angles(
            angles, extrinsic=extrinsic, zyx=zyx
        )
        self.assertAllClose(rot_vec_, rot_vec, atol=atol)

    def test_left_jacobian_translation_det(self, point, expected, atol):
        jacobian = self.space.jacobian_translation(point=point, left=True)
        result = gs.linalg.det(jacobian)
        self.assertAllClose(result, expected, atol=atol)

    def test_compose_regularize_angles_close_to_pi(self, point):
        result = self.space.compose(point, self.space.identity)
        expected = self.space.regularize(point)
        inv_expected = -expected
        self.assertTrue(
            gs.allclose(result, expected) or gs.allclose(result, inv_expected)
        )

        result = self.space.compose(self.space.identity, point)
        expected = self.space.regularize(point)
        inv_expected = -expected
        self.assertTrue(
            gs.allclose(result, expected) or gs.allclose(result, inv_expected)
        )



import pytest

import geomstats.backend as gs
from geomstats.test.random import RandomDataGenerator
from geomstats.test_cases.learning._base import BaseEstimatorTestCase


class WrappedGaussianProcessRandomDataGenerator(RandomDataGenerator):
    def __init__(self, space, prior):
        super().__init__(space)
        self.prior = prior

        self._oscillation = (1.0 / 20.0) * gs.array([-0.5, 0.0, 1.0])

    def random_X(self, n_samples):
        X = gs.linspace(0.0, 1.5 * gs.pi, n_samples)
        return gs.reshape((X - gs.mean(X)), (-1, 1))

    def random_dataset(self, n_samples):
        X = self.random_X(n_samples)

        y = self.prior(X)

        o = self.space.to_tangent(self._oscillation, base_point=y)
        s = X * gs.sin(5.0 * gs.pi * X)

        new_y = self.space.metric.exp(s * o, base_point=y)

        return X, new_y


class WrappedGaussianProcessTestCase(BaseEstimatorTestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = WrappedGaussianProcessRandomDataGenerator(
                self.estimator.space, self.estimator.prior
            )

    @pytest.mark.random
    def test_score_at_train_is_one(self, n_samples, atol):
        X, y = self.data_generator.random_dataset(n_samples)

        self.estimator.fit(X, y)

        res = self.estimator.score(X, y)
        self.assertAllClose(res, 1, atol=atol)

    @pytest.mark.random
    def test_predict_at_train_belongs(self, n_samples, atol):
        X, y = self.data_generator.random_dataset(n_samples)

        self.estimator.fit(X, y)

        y_ = self.estimator.predict(X)

        expected = gs.ones(n_samples, dtype=bool)
        res = self.estimator.space.belongs(y_, atol=atol)
        self.assertAllEqual(expected, res)

    @pytest.mark.random
    def test_predict_at_train_zero_std(self, n_samples, atol):
        X, y = self.data_generator.random_dataset(n_samples)

        self.estimator.fit(X, y)

        _, std = self.estimator.predict(X, return_tangent_std=True)

        expected = gs.zeros_like(std)
        self.assertAllClose(std, expected, atol=atol)

    @pytest.mark.random
    def test_sample_y_at_train_belongs(self, n_samples, atol):
        X, y = self.data_generator.random_dataset(n_samples)

        self.estimator.fit(X, y)
        y_ = self.estimator.sample_y(X)
        y_ = gs.reshape(gs.transpose(y_, [0, 2, 1]), (-1, y_.shape[1]))

        res = self.estimator.space.belongs(y_, atol=atol)
        expected = gs.ones(n_samples, dtype=bool)
        self.assertAllEqual(res, expected)



from geomstats.test_cases.learning._base import (
    BaseEstimatorTestCase,
    MeanEstimatorMixinsTestCase,
)


class IncrementalFrechetMeanTestCase(
    MeanEstimatorMixinsTestCase, BaseEstimatorTestCase
):
    def test_fit(self, X, expected, atol):
        estimate = self.estimator.fit(X).estimate_
        self.assertAllClose(estimate, expected, atol=atol)



"""Finite differences machinery."""

import geomstats.backend as gs


def forward_difference(array, delta=None, axis=-1):
    """Forward difference in a Euclidean space.

    Points live in R^m, but are a k-dim embedding (e.g. a curve).
    Assumes points are in correspondence for cases higher than dim=1.

    Parameters
    ----------
    array : array-like
        Values of a function.
    delta : float
        Spacing between points.
    axis : int
        Axis in which perform the difference.
        Must be given backwards.

    Returns
    -------
    forward_diff : array-like
        Shape in the specified axis reduces by one.
    """
    n = array.shape[axis]
    if delta is None:
        delta = 1 / (n - 1)

    point_ndim_slc = (slice(None),) * (abs(axis) - 1)

    slc = (..., slice(1, n)) + point_ndim_slc
    forward = array[slc]

    slc = (..., slice(0, n - 1)) + point_ndim_slc
    center = array[slc]
    return (forward - center) / delta


def centered_difference(array, delta=None, axis=-1, endpoints=False):
    """Centered difference in a Euclidean space.

    Points live in R^m, but are a k-dim embedding (e.g. a curve).
    Assumes points are in correspondence for cases higher than dim=1.

    Parameters
    ----------
    array : array-like
        Values of a function.
    delta : float
        Spacing between points.
    axis : int
        Axis in which perform the difference.
        Must be given backwards.
    endpoints : bool
        If True, endpoints are computed by backward and forward differences,
        respectively.

    Returns
    -------
    centered_diff : array-like
        Same shape as array.
    """
    n = array.shape[axis]
    if delta is None:
        delta = 1 / (n - 1)

    point_ndim_slc = (slice(None),) * (abs(axis) - 1)

    slc = (..., slice(2, n)) + point_ndim_slc
    forward = array[slc]

    slc = (..., slice(0, n - 2)) + point_ndim_slc
    backward = array[slc]
    diff = (forward - backward) / (2 * delta)

    if endpoints:
        slc_left = (..., [0]) + point_ndim_slc
        slc_left_forward = (..., [1]) + point_ndim_slc
        diff_left = (array[slc_left_forward] - array[slc_left]) / delta

        slc_right = (..., [-1]) + point_ndim_slc
        slc_right_backward = (..., [-2]) + point_ndim_slc
        diff_right = (array[slc_right] - array[slc_right_backward]) / delta

        slc_right = (..., [-1]) + point_ndim_slc
        return gs.concatenate((diff_left, diff, diff_right), axis=axis)

    return diff


def second_centered_difference(array, delta=None, axis=-1):
    """Second centered difference in a Euclidean space.

    Points live in R^m, but are a k-dim embedding (e.g. a curve).
    Assumes points are in correspondence for cases higher than dim=1.

    Parameters
    ----------
    array : array-like
        Values of a function.
    delta : float
        Spacing between points.
    axis : int
        Axis in which perform the difference.
        Must be given backwards.

    Returns
    -------
    second_centered_diff : array-like
        Shape in the specified axis reduces by two (endpoints).
    """
    n = array.shape[axis]
    if delta is None:
        delta = 1 / (n - 1)

    point_ndim_slc = (slice(None),) * (abs(axis) - 1)

    slc = (..., slice(2, n)) + point_ndim_slc
    forward = array[slc]

    slc = (..., slice(0, n - 2)) + point_ndim_slc
    backward = array[slc]

    slc = (..., slice(1, n - 1)) + point_ndim_slc
    central = array[slc]

    return (forward + backward - 2 * central) / (delta**2)



import math

import pytest

import geomstats.algebra_utils as utils
import geomstats.backend as gs
from geomstats.geometry.hypersphere import Hypersphere
from geomstats.test.test_case import TestCase


class AlgebraUtilsTestCase(TestCase):
    @pytest.mark.random
    def test_taylor_functions_even_exp(self, taylor_function, exponent, atol):
        x = 10 ** (-exponent)
        expected = taylor_function["function"](math.sqrt(x))
        result = utils.taylor_exp_even_func(x, taylor_function, order=4)
        self.assertAllClose(result, expected, atol=atol)

    def test_from_vector_to_diagonal_matrix(
        self,
        vector,
        expected,
        atol,
        num_diag=0,
    ):
        res = utils.from_vector_to_diagonal_matrix(vector, num_diag=num_diag)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_rotate_points(self, n_points, atol):
        north_pole = gs.array([1.0, 0.0, 0.0])
        sphere = Hypersphere(dim=2, equip=False)

        end_point = sphere.random_point()
        res = utils.rotate_points(north_pole, end_point)
        self.assertAllClose(res, end_point, atol=atol)

        points = sphere.random_point(n_points)
        res = utils.rotate_points(points, north_pole)
        self.assertAllClose(res, points, atol=atol)

    @pytest.mark.random
    def test_columnwise_scaling(self, n_points, m, n, atol):
        batch_shape = () if n_points == 1 else (n_points,)

        diag_vec = gs.random.uniform(size=batch_shape + (n,))
        point = gs.random.uniform(size=batch_shape + (m, n))

        res = utils.columnwise_scaling(diag_vec, point)
        res_ = gs.matmul(point, gs.vec_to_diag(diag_vec))
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_rowwise_scaling(self, n_points, m, n, atol):
        batch_shape = () if n_points == 1 else (n_points,)

        diag_vec = gs.random.uniform(size=batch_shape + (m,))
        point = gs.random.uniform(size=batch_shape + (m, n))

        res = utils.rowwise_scaling(diag_vec, point)
        res_ = gs.matmul(gs.vec_to_diag(diag_vec), point)
        self.assertAllClose(res, res_, atol=atol)



"""Class for (principal) fiber bundles.

Lead author: Nicolas Guigui.
"""

import logging
import math
from abc import ABC, abstractmethod

import geomstats.backend as gs
from geomstats.numerics.optimization import ScipyMinimize
from geomstats.vectorization import check_is_batch, get_batch_shape


def _from_base(method):
    """Decorate method in order to avoid recursive calls."""
    method._from_base = True
    return method


class AlignerAlgorithm(ABC):
    """Base class for point to point aligner.

    Parameters
    ----------
    total_space : Manifold
        Space equipped with a group action and a group-invariant metric.
    """

    def __init__(self, total_space):
        self._total_space = total_space

    @abstractmethod
    def align(self, point, base_point):
        """Align point to base point.

        Parameters
        ----------
        point : array-like, shape=[..., *total_space.shape]
            Point to align.
        base_point : array-like, shape=[..., *total_space.shape]
            Base point.

        Returns
        -------
        aligned_point : array-like, shape=[..., *total_space.shape]
            Aligned point.
        """


class DistanceMinimizingAligner(AlignerAlgorithm):
    """Aligment based on minimization of squared distance.

    Parameters
    ----------
    total_space : Manifold
        Space equipped with a group action and a group-invariant metric.
    optimizer : ScipyMinimize
        Optimizer to solve minimization problem.
    group_elem_shape : tuple
        Shape of the group element representation.
    """

    def __init__(self, total_space, optimizer=None, group_elem_shape=None):
        super().__init__(total_space)
        if optimizer is None:
            optimizer = ScipyMinimize(
                method="L-BFGS-B",
                autodiff_jac=True,
            )
        self.optimizer = optimizer

        if group_elem_shape is None:
            group_elem_shape = self._total_space.group_action.group_elem_shape
        self.group_elem_shape = group_elem_shape

    def _objective(self, point, base_point, batch_shape):
        """Objective function.

        Parameters
        ----------
        point : array-like, shape=[..., *total_space.shape]
            Point to align to base point.
        base_point : array-like, shape=[..., *total_space.shape]
            Point wrt alignment is performed.
        batch_shape : tuple
            Batch shape.
        """

        def sum_squared_dist(param):
            """Objective function.

            Parameters
            ----------
            param : array-like
                Flat representation of group element.
            """
            group_elem = gs.reshape(param, batch_shape + self.group_elem_shape)
            aligned_point = self._total_space.group_action(group_elem, point)
            return gs.sum(
                self._total_space.metric.squared_dist(aligned_point, base_point)
            )

        return sum_squared_dist

    def align(self, point, base_point):
        """Align point to base point.

        Parameters
        ----------
        point : array-like, shape=[..., *total_space.shape]
            Point to align.
        base_point : array-like, shape=[..., *total_space.shape]
            Base point.

        Returns
        -------
        aligned_point : array-like, shape=[..., *total_space.shape]
            Aligned point.
        """
        batch_shape = get_batch_shape(self._total_space.point_ndim, point, base_point)
        objective = self._objective(point, base_point, batch_shape)

        initial_param = gs.flatten(
            gs.random.rand(math.prod(batch_shape + self.group_elem_shape))
        )

        sol = self.optimizer.minimize(
            objective,
            initial_param,
        )

        group_elem = gs.reshape(sol.x, batch_shape + self.group_elem_shape)
        aligned_point = self._total_space.group_action(group_elem, point)

        return aligned_point


class AlternatingAligner(AlignerAlgorithm):
    """Alternate alignment algorithm.

    Assumes total space is equipped with several group actions.
    Aligns points wrt these group actions by alternate minimization
    wrt each of them (see e.g. [SK2016]_ for more details).

    This alignment results in an approximate quotient of the product
    group action.

    Parameters
    ----------
    total_space : Manifold
        Manifold equipped with a quotient structure.
    threshold : float
        Distance between consecutive aligned points for which
        convergence is considered reached.
    max_iter : int
        Maximum number of iterations.
    verbose : boolean
        If log number of iterations need for convergence.

    References
    ----------
    .. [SK2016] Anuj Srivastava, and Eric P. Klassen.
        Functional and Shape Data Analysis. Springer, 2016.
    """

    def __init__(self, total_space, threshold=1e-3, max_iter=20, verbose=0):
        super().__init__(total_space=total_space)

        self.threshold = threshold
        self.max_iter = max_iter
        self.verbose = verbose

        self.total_spaces = []
        for group_action in total_space.group_action:
            total_space = total_space.new(equip=True)
            total_space.equip_with_group_action(group_action)
            total_space.equip_with_quotient()
            self.total_spaces.append(total_space)

    def _align_single(self, point, base_point):
        """Align point to base point.

        Parameters
        ----------
        point : array-like, shape=[*point_shape]
            Discrete curve to align.
        base_point : array-like, shape=[*point_shape]
            Reference discrete curve.

        Returns
        -------
        aligned : array-like, shape=[*point_shape]
            Aligned point.
        """
        aligned_point = previous_aligned_point = point
        for index in range(self.max_iter):
            for total_space in self.total_spaces:
                aligned_point = total_space.fiber_bundle.align(
                    aligned_point, base_point
                )

            gap = self._total_space.metric.dist(aligned_point, previous_aligned_point)
            previous_aligned_point = aligned_point

            if gap < self.threshold:
                if self.verbose > 0:
                    logging.info(
                        f"Convergence of alignment reached after {index + 1} "
                        "iterations."
                    )

                break
        else:
            logging.warning(
                f"Maximum number of iterations {self.max_iter} reached during "
                "alignment. The result may be inaccurate."
            )
        return aligned_point

    def align(self, point, base_point):
        """Align point to base point.

        Parameters
        ----------
        point : array-like, shape=[..., *point_shape]
            Discrete curve to align.
        base_point : array-like, shape=[..., *point_shape]
            Reference discrete curve.

        Returns
        -------
        aligned : array-like, shape=[..., *point_shape]
            Aligned point.
        """
        is_batch = check_is_batch(
            self._total_space.point_ndim,
            point,
            base_point,
        )
        if not is_batch:
            return self._align_single(point, base_point)

        if point.ndim != base_point.ndim:
            point, base_point = gs.broadcast_arrays(point, base_point)

        return gs.stack(
            [
                self._align_single(point_, base_point_)
                for point_, base_point_ in zip(point, base_point)
            ]
        )


class FiberBundle:
    """Class for (principal) fiber bundles.

    This class implements abstract methods for fiber bundles, or more
    generally manifolds, with a submersion map, or a right Lie group action.

    Parameters
    ----------
    total_space : Manifold
        Space equipped with a group action and a group-invariant metric.
    aligner : AlignerAlgorithm
        If True and autodiff works, instantiates default
        DistanceMinimizationBasedAligner.
    """

    def __init__(self, total_space, aligner=None):
        self._total_space = total_space
        if aligner is True:
            if isinstance(total_space.group_action, tuple):
                aligner = AlternatingAligner(total_space)
            else:
                aligner = (
                    DistanceMinimizingAligner(total_space)
                    if gs.has_autodiff()
                    else None
                )

        self.aligner = aligner

    @staticmethod
    def riemannian_submersion(point):
        """Project a point to base manifold.

        This is the projection of the fiber bundle, defined on the total
        space, with values in the base manifold. This map is surjective.
        By default, the base manifold is not explicit but is identified with a
        local section of the fiber bundle, so the submersion is the identity
        map.

        Parameters
        ----------
        point : array-like, shape=[..., {total_space.dim, [n, m]}]
            Point of the total space.

        Returns
        -------
        projection : array-like, shape=[..., {base_dim, [n, m]}]
            Point of the base manifold.
        """
        return gs.copy(point)

    @staticmethod
    def lift(point):
        """Lift a point to total space.

        This is a section of the fiber bundle, defined on the base manifold,
        with values in the total space. It means that submersion applied after
        lift results in the identity map. By default, the base manifold
        is not explicit but is identified with a section of the fiber bundle,
        so the lift is the identity map.

        Parameters
        ----------
        point : array-like, shape=[..., {base_dim, [n, m]}]
            Point of the base manifold.

        Returns
        -------
        lift : array-like, shape=[..., {total_space.dim, [n, m]}]
            Point of the total space.
        """
        return gs.copy(point)

    @_from_base
    def tangent_riemannian_submersion(self, tangent_vec, base_point):
        """Project a tangent vector to base manifold.

        This is the differential of the projection of the fiber bundle,
        defined on the tangent space of a point of the total space,
        with values in the tangent space of the projection of this point in the
        base manifold. This map is surjective. By default, the base manifold
        is not explicit but is identified with a horizontal section of the
        fiber bundle, so the tangent submersion is the horizontal projection.

        Parameters
        ----------
        tangent_vec :  array-like, shape=[..., {total_space.dim, [n, m]}]
            Tangent vector to the total space at `base_point`.
        base_point: array-like, shape=[..., {total_space.dim, [n, m]}]
            Point of the total space.

        Returns
        -------
        projection: array-like, shape=[..., {base_dim, [n, m]}]
            Tangent vector to the base manifold.
        """
        return self.horizontal_projection(tangent_vec, base_point)

    def align(self, point, base_point):
        """Align point to base point.

        Parameters
        ----------
        point : array-like, shape=[..., *total_space.shape]
            Point to align.
        base_point : array-like, shape=[..., *total_space.shape]
            Base point.

        Returns
        -------
        aligned_point : array-like, shape=[..., *total_space.shape]
            Aligned point.
        """
        if self.aligner is None:
            raise NotImplementedError("Alignment is not implemented.")
        return self.aligner.align(point, base_point)

    @_from_base
    def horizontal_projection(self, tangent_vec, base_point):
        r"""Project to horizontal subspace.

        Compute the horizontal component of a tangent vector at a
        base point from:
            1. the vertical projection
            2. the horizontal lift of the tangent submersion

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., {total_space.dim, [n, m]}]
            Tangent vector to the total space at `base_point`.
        base_point : array-like, shape=[..., {total_space.dim, [n, m]}]
            Point on the total space.

        Returns
        -------
        horizontal : array-like, shape=[..., {total_space.dim, [n, m]}]
            Horizontal component of `tangent_vec`.
        """
        if not hasattr(self.vertical_projection, "_from_base"):
            ver_tangent_vec = self.vertical_projection(tangent_vec, base_point)
            return tangent_vec - ver_tangent_vec

        if not (
            hasattr(self.horizontal_lift, "_from_base")
            or hasattr(self.tangent_riemannian_submersion, "_from_base")
        ):
            return self.horizontal_lift(
                self.tangent_riemannian_submersion(tangent_vec, base_point),
                fiber_point=base_point,
            )

        raise NotImplementedError("Horizontal projection is not implemented.")

    @_from_base
    def vertical_projection(self, tangent_vec, base_point):
        r"""Project to vertical subspace.

        Compute the vertical component of a tangent vector :math:`w` at a
        base point :math:`P` by removing the horizontal component.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., {total_space.dim, [n, m]}]
            Tangent vector to the total space at `base_point`.
        base_point : array-like, shape=[..., {total_space.dim, [n, m]}]
            Point on the total space.

        Returns
        -------
        vertical : array-like, shape=[..., {total_space.dim, [n, m]}]
            Vertical component of `tangent_vec`.
        """
        return tangent_vec - self.horizontal_projection(tangent_vec, base_point)

    def is_horizontal(self, tangent_vec, base_point, atol=gs.atol):
        """Evaluate if the tangent vector is horizontal at base_point.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., {total_space.dim, [n, m]}]
            Tangent vector.
        base_point : array-like, shape=[..., {total_space.dim, [n, m]}]
            Point on the manifold.
            Optional, default: None.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol

        Returns
        -------
        is_horizontal : bool
            Boolean denoting if tangent vector is horizontal.
        """
        return gs.all(
            gs.isclose(
                tangent_vec,
                self.horizontal_projection(tangent_vec, base_point),
                atol=atol,
            ),
            axis=(-2, -1),
        )

    def is_vertical(self, tangent_vec, base_point, atol=gs.atol):
        """Evaluate if the tangent vector is vertical at base_point.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., {total_space.dim, [n, m]}]
            Tangent vector.
        base_point : array-like, shape=[..., {total_space.dim, [n, m]}]
            Point on the manifold.
            Optional, default: None.
        atol : float
            Absolute tolerance.
            Optional, default: backend atol.

        Returns
        -------
        is_vertical : bool
            Boolean denoting if tangent vector is vertical.
        """
        return gs.all(
            gs.isclose(
                0.0,
                self.tangent_riemannian_submersion(tangent_vec, base_point),
                atol=atol,
            ),
            axis=(-2, -1),
        )

    @_from_base
    def horizontal_lift(self, tangent_vec, base_point=None, fiber_point=None):
        """Lift a tangent vector to a horizontal vector in the total space.

        It means that horizontal lift is the inverse of the restriction of the
        tangent submersion to the horizontal space at point, where point must
        be in the fiber above the base point. By default, the base manifold
        is not explicit but is identified with a horizontal section of the
        fiber bundle, so the horizontal lift is the horizontal projection.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., {base_dim, [n, m]}]
        fiber_point : array-like, shape=[..., {ambient_dim, [n, m]}]
            Point of the total space.
            Optional, default : None. The `lift` method is used to compute a
            point at which to compute a tangent vector.
        base_point : array-like, shape=[..., {base_dim, [n, m]}]
            Point of the base space.
            Optional, default : None. In this case, point must be given,
            and `submersion` is used to compute the base_point if needed.

        Returns
        -------
        horizontal_lift : array-like, shape=[..., {total_space.dim, [n, m]}]
            Horizontal tangent vector to the total space at point.
        """
        if base_point is None and fiber_point is None:
            raise ValueError(
                "Either a point (of the total space) or a "
                "base point (of the base manifold) must be "
                "given."
            )

        if fiber_point is None:
            fiber_point = self.lift(base_point)

        return self.horizontal_projection(tangent_vec, fiber_point)

    def integrability_tensor(self, tangent_vec_a, tangent_vec_b, base_point):
        r"""Compute the fundamental tensor A of the submersion.

        The fundamental integrability tensor A is defined for tangent vectors
        :math:`X = tangent\_vec\_a` and :math:`Y = tangent\_vec\_b` of the
        total space by [ONeill]_ as
        :math:`A_X Y = ver\nabla_{hor X} (hor Y) + hor \nabla_{hor X}( ver Y)`
        where :math:`hor, ver` are the horizontal and vertical projections
        and :math:`\nabla` is the connection of the total space.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., {total_space.dim, [n, m]}]
            Tangent vector at `base_point`.
        tangent_vec_b : array-like, shape=[..., {total_space.dim, [n, m]}]
            Tangent vector at `base_point`.
        base_point : array-like, shape=[..., {total_space.dim, [n, m]}]
            Point of the total space.

        Returns
        -------
        vector : array-like, shape=[..., {total_space.dim, [n, m]}]
            Tangent vector at `base_point`, result of the A tensor applied to
            `tangent_vec_a` and `tangent_vec_b`.

        References
        ----------
        .. [ONeill]  O’Neill, Barrett. The Fundamental Equations of a
            Submersion, Michigan Mathematical Journal 13, no. 4
            (December 1966): 459–69. https://doi.org/10.1307/mmj/1028999604.
        """
        raise NotImplementedError

    def integrability_tensor_derivative(
        self,
        horizontal_vec_x,
        horizontal_vec_y,
        nabla_x_y,
        tangent_vec_e,
        nabla_x_e,
        base_point,
    ):
        r"""Compute the covariant derivative of the integrability tensor A.

        The covariant derivative :math:`\nabla_X (A_Y E)` in total space is
        necessary to compute the covariant derivative of the directional
        curvature in a submersion. The components :math:`\nabla_X (A_Y E)`
        and :math:`A_Y E` are computed at base-point :math:`P = base\_point`
        for horizontal vector fields :math:`X, Y` extending the values
        given in argument :math:`X|_P = horizontal\_vec\_x`,
        :math:`Y|_P = horizontal\_vec\_y` and a general vector field
        :math:`E` extending :math:`E|_x = tangent\_vec\_e`
        in a neighborhood of x with covariant derivatives
        :math:`\nabla_X Y |_P = nabla_x y` and
        :math:`\nabla_X E |_P = nabla_x e`.

        Parameters
        ----------
        horizontal_vec_x : array-like, shape=[..., {total_space.dim, [n, m]}]
            Horizontal tangent vector at `base_point`.
        horizontal_vec_y : array-like, shape=[..., {total_space.dim, [n, m]}]
            Horizontal tangent vector at `base_point`.
        nabla_x_y : array-like, shape=[..., {total_space.dim, [n, m]}]
            Tangent vector at `base_point`.
        tangent_vec_e : array-like, shape=[..., {total_space.dim, [n, m]}]
            Tangent vector at `base_point`.
        nabla_x_e : array-like, shape=[..., {total_space.dim, [n, m]}]
            Tangent vector at `base_point`.
        base_point : array-like, shape=[..., {total_space.dim, [n, m]}]
            Point of the total space.

        Returns
        -------
        nabla_x_a_y_e : array-like, shape=[..., {total_space.dim, [n, m]}]
            Tangent vector at `base_point`, result of :math:`\nabla_X
            (A_Y E)`.
        a_y_e : array-like, shape=[..., {ambient_dim, [n, n]}]
            Tangent vector at `base_point`, result of :math:`A_Y E`.

        References
        ----------
        .. [Pennec] Pennec, Xavier. Computing the curvature and its gradient
            in Kendall shape spaces. Unpublished.
        """
        raise NotImplementedError



from geomstats.test_cases.learning._base import BaseEstimatorTestCase


class KNearestNeighborsClassifierTestCase(BaseEstimatorTestCase):
    def test_predict(self, X_train, y_train, X_test, y_test):
        self.estimator.fit(X_train, y_train)

        y_pred = self.estimator.predict(X_test)
        self.assertAllEqual(y_pred, y_test)

    def test_predict_proba(self, X_train, y_train, X_test, expected_proba, atol):
        self.estimator.fit(X_train, y_train)

        proba = self.estimator.predict_proba(X_test)
        self.assertAllClose(proba, expected_proba, atol=atol)



import pytest

import geomstats.backend as gs
from geomstats.test.test_case import TestCase


class LogNormalTestCase(TestCase):
    @pytest.mark.random
    def test_sample_belongs(self, n_samples, atol):
        samples = self.distribution.sample(n_samples)

        res = self.distribution.space.belongs(samples, atol=atol)
        expected = gs.ones(n_samples, dtype=bool)
        self.assertAllEqual(res, expected)



r"""Wrapped Gaussian Process.

Lead author: Arthur Pignet

Extension of Gaussian Processes to Riemannian Manifolds,
introduced in [Mallasto]_.

References
----------
.. [Mallasto] Mallasto, A. and Feragen, A.
    “Wrapped gaussian process regression on riemannian manifolds.”
    IEEE/CVF Conference on Computer Vision and Pattern Recognition
    (2018)

"""

from sklearn.base import BaseEstimator, MultiOutputMixin
from sklearn.gaussian_process import GaussianProcessRegressor

import geomstats.backend as gs
from geomstats.learning._sklearn import RegressorMixin


class WrappedGaussianProcess(MultiOutputMixin, RegressorMixin, BaseEstimator):
    r"""Wrapped Gaussian Process.

    The implementation is based on the algorithm 4 of [1]_.

    Parameters
    ----------
    space : Manifold
        Equipped manifold.
    prior : callable
        Associate to each input a manifold valued point.

    References
    ----------
    .. [1] Mallasto, A. and Feragen, A. Wrapped gaussian process
        regression on riemannian manifolds. In 2018 IEEE/CVF
        Conference on Computer Vision and Pattern Recognition
    """

    def __init__(self, space, prior):
        self.space = space
        self.prior = prior

        self.euclidean_gpr = GaussianProcessRegressor(
            kernel=None,
            alpha=1e-10,
            optimizer="fmin_l_bfgs_b",
            n_restarts_optimizer=0,
            normalize_y=False,
            copy_X_train=True,
            random_state=None,
        )

        self.tangent_y_train_ = None

    def set(self, **kwargs):
        """Set euclidean_gpr parameters.

        Especially useful for one line instantiations.
        """
        for param_name, value in kwargs.items():
            if not hasattr(self.euclidean_gpr, param_name):
                raise ValueError(f"Unknown parameter {param_name}.")

            setattr(self.euclidean_gpr, param_name, value)
        return self

    def _get_tangent_targets(self, X, y):
        """Compute the tangent targets, using the provided prior.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features) or list of object
            Feature vectors or other representations of training data.
        y : array-like of shape (n_samples,) or (n_samples, n_targets)
        or (n_samples, n1_targets, n2_targets) for
        matrix-valued targets.
            Target values. The target must belongs to the manifold space

        Returns
        -------
        tangent_y : array-like of shape (n_samples,) or (n_samples, n_targets)
        or (n_samples, n1_targets, n2_targets)
                Target projected on the associated (by the prior) tangent space.
        """
        base_points = self.prior(X)
        return self.space.metric.log(y, base_point=base_points)

    def fit(self, X, y):
        """Fit Wrapped Gaussian process regression model.

        The Wrapped Gaussian process is fit through the following steps:

        - Compute the tangent dataset using the prior
        - Fit a Gaussian process regression on the tangent dataset
        - Store the resulting euclidean Gaussian process

        Parameters
        ----------
        X : array-like, shape=[n_samples,]
            Training input samples.
        y : array-like, shape[n_samples, {dim, [n,n]}]
            Training target values.

        Returns
        -------
        self : object
            Returns self.
        """
        self.tangent_y_train_ = tangent_y = self._get_tangent_targets(X, y)
        tangent_y = gs.reshape(tangent_y, (y.shape[0], -1))

        self.euclidean_gpr.fit(X, tangent_y)

        return self

    def predict(self, X, return_tangent_std=False, return_tangent_cov=False):
        """Predict using the Gaussian process regression model.

        A fitted Wrapped Gaussian process can be use to predict values
        through the following steps:

        - Use the stored Gaussian process regression on the dataset to
          return tangent predictions
        - Compute the base-points using the prior
        - Map the tangent predictions on the manifold via the metric's exp
          with the base-points yielded by the prior

        We can also predict based on an unfitted model by using the GP prior.
        In addition to the mean of the predictive distribution, optionally also
        returns its standard deviation (`return_std=True`) or covariance
        (`return_cov=True`). Note that at most one of the two can be requested.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features) or list of object
            Query points where the GP is evaluated.
        return_tangent_std : bool, default=False
            If True, the standard-deviation of the predictive distribution on at
            the query points in the tangent space is returned along with the mean.
        return_tangent_cov : bool, default=False
            If True, the covariance of the joint predictive distribution at
            the query points in the tangent space is returned along with the mean.

        Returns
        -------
        y_mean : ndarray of shape (n_samples,) or (n_samples, n_targets)
            Mean of predictive distribution a query points.
        y_std : ndarray of shape (n_samples,) or (n_samples, n_targets), optional
            Standard deviation of predictive distribution at query points in
            the tangent space.
            Only returned when `return_std` is True.
        y_cov : ndarray of shape (n_samples, n_samples) or \
                (n_samples, n_samples, n_targets), optional
            Covariance of joint predictive distribution a query points
            in the tangent space.
            Only returned when `return_cov` is True.
            In the case where the target is matrix valued,
            return the covariance of the vectorized prediction.
        """
        euc_result = self.euclidean_gpr.predict(
            X, return_cov=return_tangent_cov, return_std=return_tangent_std
        )

        return_multiple = return_tangent_std or return_tangent_cov
        tangent_means = gs.from_numpy(euc_result[0] if return_multiple else euc_result)

        base_points = self.prior(X)
        tangent_means = gs.reshape(
            tangent_means,
            (X.shape[0], *self.space.shape),
        )
        y_mean = self.space.metric.exp(tangent_means, base_point=base_points)

        if return_multiple:
            tangent_std_cov = gs.from_numpy(euc_result[1])
            return (y_mean, tangent_std_cov)

        return y_mean

    def sample_y(self, X, n_samples=1, random_state=0):
        """Draw samples from Wrapped Gaussian process and evaluate at X.

        A fitted Wrapped Gaussian process can be use to sample
        values through the following steps:

        - Use the stored Gaussian process regression on the dataset
          to sample tangent values
        - Compute the base-points using the prior
        - Flatten (and repeat if needed) both the base-points and the
          tangent samples to benefit from vectorized computation.
        - Map the tangent samples on the manifold via the metric's exp with the
          flattened and repeated base-points yielded by the prior

        Parameters
        ----------
        X : array-like of shape (n_samples_X, n_features) or list of object
            Query points where the WGP is evaluated.
        n_samples : int, default=1
            Number of samples drawn from the Wrapped Gaussian process per query
            point.
        random_state : int, RandomState instance or None, default=0
            Determines random number generation to randomly draw samples.
            Pass an int for reproducible results across multiple function
            calls.

        Returns
        -------
        y_samples : ndarray of shape (n_samples_X, n_samples), or \
            (n_samples_X, *target_shape, n_samples)
            Values of n_samples samples drawn from wrapped Gaussian process and
            evaluated at query points.
        """
        tangent_samples = gs.from_numpy(
            self.euclidean_gpr.sample_y(X, n_samples, random_state)
        )

        if gs.ndim(tangent_samples) > 2:
            tangent_samples = gs.moveaxis(tangent_samples, -2, -1)

        flat_tangent_samples = gs.reshape(tangent_samples, (-1, *self.space.shape))

        base_points = gs.repeat(self.prior(X), n_samples, axis=0)

        flat_y_samples = self.space.metric.exp(
            flat_tangent_samples, base_point=base_points
        )

        y_samples = gs.reshape(
            flat_y_samples, (X.shape[0], n_samples, *self.space.shape)
        )

        if gs.ndim(tangent_samples) > 2:
            y_samples = gs.moveaxis(y_samples, 1, -1)

        return y_samples



import pytest

import geomstats.backend as gs
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.vectorization import get_batch_shape


class InformationManifoldMixinTestCase(TestCase):
    @pytest.mark.shape
    def test_sample_shape(self, n_points, n_samples):
        point = self.data_generator.random_point(n_points)

        res = self.space.sample(point, n_samples=n_samples)

        expected_shape = (
            get_batch_shape(self.space.point_ndim, point)
            + (n_samples,)
            + self.space.support_shape
        )
        self.assertEqual(res.shape, expected_shape)

    def _check_sample_belongs_to_support(self, sample, atol):
        raise NotImplementedError("Need to define `_check_sample_belongs_to_support`.")

    @pytest.mark.random
    def test_sample_belongs_to_support(self, n_points, n_samples, atol):
        point = self.data_generator.random_point(n_points)
        sample = self.space.sample(point, n_samples)

        self._check_sample_belongs_to_support(sample, atol)

    def test_point_to_pdf(self, x, point, expected, atol):
        pdf = self.space.point_to_pdf(point)
        res = pdf(x)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_point_to_pdf_vec(self, n_reps, n_samples, atol):
        point = self.data_generator.random_point()

        x = self.space.sample(point, n_samples)
        expected = self.space.point_to_pdf(point)(x)

        vec_data = generate_vectorization_data(
            data=[dict(x=x, point=point, expected=expected, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_point_to_pdf_against_scipy(self, n_points, n_samples, atol):
        point = self.data_generator.random_point(n_points)

        sample_point = point if n_points == 1 else point[0]
        sample = self.space.sample(sample_point, n_samples=n_samples)
        print(sample.shape)

        res = self.space.point_to_pdf(point)(sample)
        res_ = self.random_variable.pdf(sample, point)

        self.assertAllClose(res, res_, atol=atol)

    def test_point_to_cdf(self, x, point, expected, atol):
        cdf = self.space.point_to_cdf(point)
        res = cdf(x)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_point_to_cdf_vec(self, n_reps, n_samples, atol):
        point = self.data_generator.random_point()

        x = self.space.sample(point, n_samples)
        expected = self.space.point_to_cdf(point)(x)

        vec_data = generate_vectorization_data(
            data=[dict(point=point, expected=expected, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_point_to_cdf_bounds(self, n_points, n_samples):
        point = self.data_generator.random_point(n_points)
        pdf = self.space.point_to_cdf(point)

        sample_point = point if n_points == 1 else point[0]
        sample = self.space.sample(sample_point, n_samples)
        res = pdf(sample)
        self.assertTrue(gs.all(res >= 0.0))
        self.assertTrue(gs.all(res <= 1.0))



"""Euclidean space."""

import math

import geomstats.backend as gs
from geomstats.geometry.base import VectorSpace
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.vectorization import check_is_batch, repeat_out


class Euclidean(VectorSpace):
    """Class for Euclidean spaces.

    By definition, a Euclidean space is a vector space of a given
    dimension, equipped with a Euclidean metric.

    Parameters
    ----------
    dim : int
        Dimension of the Euclidean space.
    """

    def __init__(self, dim, equip=True):
        super().__init__(
            dim=dim,
            shape=(dim,),
            equip=equip,
        )

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return CanonicalEuclideanMetric

    @property
    def identity(self):
        """Identity of the group.

        Returns
        -------
        identity : array-like, shape=[n]
        """
        return gs.zeros(self.dim)

    def _create_basis(self):
        """Create the canonical basis."""
        return gs.eye(self.dim)

    def exp(self, tangent_vec, base_point):
        """Compute the group exponential, which is simply the addition.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n]
            Tangent vector at base point.
        base_point : array-like, shape=[..., n]
            Point from which the exponential is computed.

        Returns
        -------
        point : array-like, shape=[..., n]
            Group exponential.
        """
        return tangent_vec + base_point


class EuclideanMetric(RiemannianMetric):
    """Class for a Euclidean metric.

    This metric is:
    - flat: the inner-product is independent of the base point;
    - positive definite
    """

    def __init__(self, space, metric_matrix=None, signature=None):
        super().__init__(space, signature=signature)
        self._check_metric_matrix_dim(space, metric_matrix)
        if metric_matrix is None and space.point_ndim == 1:
            metric_matrix = gs.eye(space.dim)

        self.metric_matrix_ = metric_matrix

    @staticmethod
    def _check_metric_matrix_dim(space, metric_matrix):
        """Check metric matrix dimension."""
        if metric_matrix is None:
            return

        expected_shape = (space.dim, space.dim)
        if metric_matrix.shape != expected_shape:
            raise ValueError(
                f"metric_matrix shape is {metric_matrix.shape};"
                f"expected: {expected_shape}"
            )

    def metric_matrix(self, base_point=None):
        """Compute the inner-product matrix, independent of the base point.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_prod_mat : array-like, shape=[..., dim, dim]
            Inner-product matrix.
        """
        if self._space.point_ndim > 1:
            raise NotImplementedError("`metric_matrix` is not implemented.")

        dim = self._space.dim
        return repeat_out(
            self._space.point_ndim,
            gs.copy(self.metric_matrix_),
            base_point,
            out_shape=(dim, dim),
        )

    def inner_product_derivative_matrix(self, base_point=None):
        r"""Compute derivative of the inner prod matrix at base point.

        Writing :math:`g_{ij}` the inner-product matrix at base point,
        this computes :math:`mat_{ijk} = \partial_k g_{ij}`, where the
        index k of the derivation is put last.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        metric_derivative : array-like, shape=[..., dim, dim, dim]
            Derivative of the inner-product matrix, where the index
            k of the derivation is last: :math:`mat_{ijk} = \partial_k g_{ij}`.
        """
        if self._space.point_ndim > 1:
            raise NotImplementedError(
                "`inner_product_derivative_matrix` is not implemented."
            )

        dim = self._space.dim
        shape = (dim, dim, dim)
        return repeat_out(
            self._space.point_ndim, gs.zeros(shape), base_point, out_shape=shape
        )

    def christoffels(self, base_point=None):
        """Christoffel symbols associated with the connection.

        The contravariant index is on the first dimension.

        Parameters
        ----------
        base_point : array-like, shape=[..., dim]
            Point on the manifold.

        Returns
        -------
        gamma : array-like, shape=[..., dim, dim, dim]
            Christoffel symbols, with the contravariant index on
            the first dimension.
        """
        if self._space.point_ndim > 1:
            raise NotImplementedError("The Christoffel symbols are not implemented.")

        dim = self._space.dim
        shape = (dim, dim, dim)
        gamma = gs.zeros(shape)
        return repeat_out(self._space.point_ndim, gamma, base_point, out_shape=shape)

    def exp(self, tangent_vec, base_point):
        """Compute exp map of a base point in tangent vector direction.

        The Riemannian exponential is vector addition in the Euclidean space.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at base point.
        base_point : array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        exp : array-like, shape=[..., dim]
            Riemannian exponential.
        """
        return base_point + tangent_vec

    def log(self, point, base_point):
        """Compute log map using a base point and other point.

        The Riemannian logarithm is the subtraction in the Euclidean space.

        Parameters
        ----------
        point: array-like, shape=[..., dim]
            Point.
        base_point: array-like, shape=[..., dim]
            Base point.

        Returns
        -------
        log: array-like, shape=[..., dim]
            Riemannian logarithm.
        """
        return point - base_point

    def parallel_transport(
        self, tangent_vec, base_point=None, direction=None, end_point=None
    ):
        r"""Compute the parallel transport of a tangent vector.

        On a Euclidean space, the parallel transport of a (tangent) vector
        returns the vector itself.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., dim]
            Tangent vector at base point to be transported.
        base_point : array-like, shape=[..., dim]
            Point on the manifold. Point to transport from.
            Optional, default: None
        direction : array-like, shape=[..., dim]
            Tangent vector at base point, along which the parallel transport
            is computed.
            Optional, default: None.
        end_point : array-like, shape=[..., dim]
            Point on the manifold. Point to transport to.
            Optional, default: None.

        Returns
        -------
        transported_tangent_vec: array-like, shape=[..., dim]
            Transported tangent vector at `exp_(base_point)(tangent_vec_b)`.
        """
        transported_tangent_vec = gs.copy(tangent_vec)
        return repeat_out(
            self._space.point_ndim,
            transported_tangent_vec,
            tangent_vec,
            base_point,
            direction,
            end_point,
            out_shape=self._space.shape,
        )

    def geodesic(self, initial_point, end_point=None, initial_tangent_vec=None):
        """Generate parameterized function for the geodesic curve.

        Geodesic curve defined by either:

        - an initial point and an initial tangent vector,
        - an initial point and an end point.

        Parameters
        ----------
        initial_point : array-like, shape=[..., dim]
            Point on the manifold, initial point of the geodesic.
        end_point : array-like, shape=[..., dim], optional
            Point on the manifold, end point of the geodesic. If None,
            an initial tangent vector must be given.
        initial_tangent_vec : array-like, shape=[..., dim],
            Tangent vector at base point, the initial speed of the geodesics.
            Optional, default: None.
            If None, an end point must be given and a logarithm is computed.

        Returns
        -------
        path : callable
            Time parameterized geodesic curve. If a batch of initial
            conditions is passed, the output array's first dimension
            represents the different initial conditions, and the second
            corresponds to time.
        """
        if end_point is None and initial_tangent_vec is None:
            raise ValueError(
                "Specify an end point or an initial tangent "
                "vector to define the geodesic."
            )
        if end_point is not None:
            if initial_tangent_vec is not None:
                raise ValueError(
                    "Cannot specify both an end point and an initial tangent vector."
                )

            initial_tangent_vec = self.log(end_point, initial_point)

        is_batch = check_is_batch(
            self._space.point_ndim, initial_point, initial_tangent_vec
        )
        if is_batch:
            initial_point = gs.expand_dims(
                initial_point, axis=-(self._space.point_ndim + 1)
            )

        ijk = "ijk"[: self._space.point_ndim]

        def path(t):
            """Generate parameterized function for geodesic curve.

            Parameters
            ----------
            t : array-like, shape=[n_points,]
                Times at which to compute points of the geodesics.
            """
            t = gs.array(t)
            t = gs.cast(t, initial_tangent_vec.dtype)
            t = gs.to_ndarray(t, to_ndim=1)
            tangent_vecs = gs.einsum(f"n,...{ijk}->...n{ijk}", t, initial_tangent_vec)
            return initial_point + tangent_vecs

        return path

    def injectivity_radius(self, base_point=None):
        """Compute the radius of the injectivity domain.

        This is is the supremum of radii r for which the exponential map is a
        diffeomorphism from the open ball of radius r centered at the base
        point onto its image.

        Parameters
        ----------
        base_point : array-like, shape=[..., {dim, [n, m]}]
            Point on the manifold.

        Returns
        -------
        radius : array-like, shape=[...,]
            Injectivity radius.
        """
        radius = gs.array(math.inf)
        return repeat_out(self._space.point_ndim, radius, base_point)


class CanonicalEuclideanMetric(EuclideanMetric):
    """Class for the canonical Euclidean metric.

    Notes
    -----
    Metric matrix is identity (NB: `EuclideanMetric` allows
    to use a different metric matrix).
    """

    def __init__(self, space):
        super().__init__(space)

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Inner product between two tangent vectors at a base point.

        Parameters
        ----------
        tangent_vec_a: array-like, shape=[..., dim]
            Tangent vector at base point.
        tangent_vec_b: array-like, shape=[..., dim]
            Tangent vector at base point.
        base_point: array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_product : array-like, shape=[...,]
            Inner-product.
        """
        inner_product = gs.dot(tangent_vec_a, tangent_vec_b)
        return repeat_out(
            self._space.point_ndim,
            inner_product,
            tangent_vec_a,
            tangent_vec_b,
            base_point,
        )

    def squared_norm(self, vector, base_point=None):
        """Compute the square of the norm of a vector.

        Squared norm of a vector associated to the inner product
        at the tangent space at a base point.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        sq_norm : array-like, shape=[...,]
            Squared norm.
        """
        sq_norm = gs.linalg.norm(vector, axis=-1) ** 2
        return repeat_out(self._space.point_ndim, sq_norm, vector, base_point)

    def norm(self, vector, base_point=None):
        """Compute norm of a vector.

        Norm of a vector associated to the inner product
        at the tangent space at a base point.

        Note: This only works for positive-definite
        Riemannian metrics and inner products.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        norm : array-like, shape=[...,]
            Norm.
        """
        norm = gs.linalg.norm(vector, axis=-1)
        return repeat_out(self._space.point_ndim, norm, vector, base_point)



r"""
Manifold of linear subspaces.

The Grassmannian :math:`Gr(n, k)` is the manifold of k-dimensional
subspaces in n-dimensional Euclidean space.

Lead author: Olivier Peltre.

:math:`Gr(n, p)` is represented by
:math:`n \times n` matrices
of rank :math:`p`  satisfying :math:`P^2 = P` and :math:`P^T = P`.
Each :math:`P \in Gr(n, p)` is identified with the unique
orthogonal projector onto :math:`{\rm Im}(P)`.

:math:`Gr(n, p)` is a homogoneous space, quotient of the special orthogonal
group by the subgroup of rotations stabilising a :math:`p`-dimensional subspace:

.. math::

    Gr(n, p) \simeq \frac {SO(n)} {SO(p) \times SO(n-p)}

It is therefore customary to represent the Grassmannian
by equivalence classes of orthogonal :math:`p`-frames in :math:`{\mathbb R}^n`.
For such a representation, work in the Stiefel manifold instead.

.. math::

    Gr(n, p) \simeq St(n, p) / SO(p)

References
----------
.. [Batzies15] Batzies, E., K. Hüper, L. Machado, and F. Silva Leite.
    “Geometric Mean and Geodesic Regression on Grassmannians.”
    Linear Algebra and Its Applications 466 (February 1, 2015):
    83–101. https://doi.org/10.1016/j.laa.2014.10.003.
"""

import geomstats.backend as gs
import geomstats.errors
from geomstats.geometry.base import LevelSet
from geomstats.geometry.general_linear import GeneralLinear
from geomstats.geometry.matrices import Matrices
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.geometry.symmetric_matrices import SymmetricMatrices
from geomstats.vectorization import repeat_out


class Grassmannian(LevelSet):
    """Class for Grassmann manifolds :math:`Gr(n, p)`.

    Parameters
    ----------
    n : int
        Dimension of the Euclidean space.
    p : int
        Dimension of the subspaces.
    """

    def __init__(self, n, p, equip=True):
        geomstats.errors.check_integer(p, "p")
        geomstats.errors.check_integer(n, "n")
        if p > n:
            raise ValueError(
                "p < n is required: p-dimensional subspaces in n dimensions."
            )

        self.n = n
        self.p = p

        dim = int(p * (n - p))
        super().__init__(dim=dim, equip=equip)

    def _define_embedding_space(self):
        return SymmetricMatrices(self.n)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return GrassmannianCanonicalMetric

    def submersion(self, point):
        r"""Submersion that defines the Grassmann manifold.

        The Grassmann manifold is defined here as embedded in the set of
        symmetric matrices, as the pre-image of the function defined around the
        projector on the space spanned by the first :math:`p` columns of the identity
        matrix by (see Exercise E.25 in [Pau07]_).

        .. math::

            \begin{pmatrix} I_p + A & B^T \\ B & D \end{pmatrix} \mapsto
                (D - B(I_p + A)^{-1}B^T, A + A^2 + B^TB)

        This map is a submersion and its zero space is the set of orthogonal
        rank-:math:`p` projectors.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]

        Returns
        -------
        submersed_point : array-like, shape=[..., n, n]

        References
        ----------
        .. [Pau07] Paulin, Frédéric. “Géométrie diﬀérentielle élémentaire,” 2007.
            https://www.imo.universite-paris-saclay.fr/~paulin/notescours/
            cours_geodiff.pdf.
        """
        p = self.p

        _, eigvecs = gs.linalg.eigh(point)
        eigvecs = gs.flip(eigvecs, -1)
        flipped_point = Matrices.mul(Matrices.transpose(eigvecs), point, eigvecs)
        b = flipped_point[..., p:, :p]
        d = flipped_point[..., p:, p:]
        a = flipped_point[..., :p, :p] - gs.eye(p)
        first = d - Matrices.mul(
            b, GeneralLinear.inverse(a + gs.eye(p)), Matrices.transpose(b)
        )
        second = a + Matrices.mul(a, a) + Matrices.mul(Matrices.transpose(b), b)
        row_1 = gs.concatenate([first, gs.zeros_like(b)], axis=-1)
        row_2 = gs.concatenate([Matrices.transpose(gs.zeros_like(b)), second], axis=-1)
        return gs.concatenate([row_1, row_2], axis=-2)

    def tangent_submersion(self, vector, point):
        """Tangent submersion.

        Parameters
        ----------
        vector : array-like, shape=[..., n, n]
        point : array-like, shape=[..., n, n]

        Returns
        -------
        submersed_vector : array-like, shape=[..., n, n]
        """
        return 2 * Matrices.to_symmetric(Matrices.mul(point, vector)) - vector

    def random_uniform(self, n_samples=1):
        r"""Sample random points from a uniform distribution.

        Following [Chikuse03]_, :math:`n\_samples * n * p` scalars are sampled
        from a standard normal distribution and reshaped to matrices,
        the projectors on their first :math:`p` columns follow a uniform
        distribution.

        Parameters
        ----------
        n_samples : int
            The number of points to sample
            Optional. default: 1.

        Returns
        -------
        projectors : array-like, shape=[..., n, n]
            Points following a uniform distribution.

        References
        ----------
        .. [Chikuse03] Yasuko Chikuse, Statistics on special manifolds,
            New York: Springer-Verlag. 2003, 10.1007/978-0-387-21540-2
        """
        points = gs.random.normal(size=(n_samples, self.n, self.p))
        full_rank = Matrices.mul(Matrices.transpose(points), points)
        projector = Matrices.mul(
            points, GeneralLinear.inverse(full_rank), Matrices.transpose(points)
        )
        return projector[0] if n_samples == 1 else projector

    def random_point(self, n_samples=1, bound=1.0):
        r"""Sample random points from a uniform distribution.

        Following [Chikuse03]_, :math:`n\_samples * n * p` scalars are sampled
        from a standard normal distribution and reshaped to matrices,
        the projectors on their first :math:`p` columns follow a uniform
        distribution.

        Parameters
        ----------
        n_samples : int
            The number of points to sample
            Optional. default: 1.

        Returns
        -------
        projectors : array-like, shape=[..., n, n]
            Points following a uniform distribution.

        References
        ----------
        .. [Chikuse03] Yasuko Chikuse, Statistics on special manifolds,
            New York: Springer-Verlag. 2003, 10.1007/978-0-387-21540-2
        """
        return self.random_uniform(n_samples)

    def to_tangent(self, vector, base_point):
        """Project a vector to a tangent space of the manifold.

        Compute the bracket (commutator) of the base_point with
        the skew-symmetric part of vector.

        Parameters
        ----------
        vector : array-like, shape=[..., n, n]
            Vector.
        base_point : array-like, shape=[..., n, n]
            Point on the manifold.

        Returns
        -------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
        """
        sym = Matrices.to_symmetric(vector)
        return Matrices.bracket(base_point, Matrices.bracket(base_point, sym))

    def projection(self, point):
        """Project a matrix to the Grassmann manifold.

        An eigenvalue decomposition of (the symmetric part of) point is used.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Point in embedding manifold.

        Returns
        -------
        projected : array-like, shape=[..., n, n]
            Projected point.
        """
        mat = Matrices.to_symmetric(point)
        _, eigvecs = gs.linalg.eigh(mat)
        diagonal = gs.array([0.0] * (self.n - self.p) + [1.0] * self.p)
        p_d = gs.einsum("...ij,...j->...ij", eigvecs, diagonal)
        return Matrices.mul(p_d, Matrices.transpose(eigvecs))


class GrassmannianCanonicalMetric(RiemannianMetric):
    """Canonical metric of the Grassmann manifold."""

    def __init__(self, space):
        super().__init__(space=space, signature=(space.dim, 0, 0))
        self._general_linear = GeneralLinear(space.n, equip=False)

    def exp(self, tangent_vec, base_point):
        """Exponentiate the invariant vector field v from base point p.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point.
            `tangent_vec` is the bracket of a skew-symmetric matrix with the
            base_point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        exp : array-like, shape=[..., n, n]
            Riemannian exponential.
        """
        expm = gs.linalg.expm
        mul = Matrices.mul
        rot = Matrices.bracket(base_point, -tangent_vec)
        return mul(expm(rot), base_point, expm(-rot))

    def log(self, point, base_point):
        r"""Compute the Riemannian logarithm of point w.r.t. base_point.

        Given :math:`P, P'` in :math:`Gr(n, p)` the logarithm from :math:`P`
        to :math:`P` is induced by the infinitesimal rotation [Batzies2015]_:

        .. math::

            Y = \frac 1 2 \log \big((2 P' - 1)(2 P - 1)\big)

        The tangent vector :math:`X` at :math:`P`
        is then recovered by :math:`X = [Y, P]`.

        Parameters
        ----------
        point : array-like, shape=[..., n, n]
            Point.
        base_point : array-like, shape=[..., n, n]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., n, n]
            Riemannian logarithm, a tangent vector at `base_point`.

        References
        ----------
        .. [Batzies2015] Batzies, Hüper, Machado, Leite.
            "Geometric Mean and Geodesic Regression on Grassmannians"
            Linear Algebra and its Applications, 466, 83-101, 2015.
        """
        id_n = self._general_linear.identity
        id_n, point, base_point = gs.convert_to_wider_dtype([id_n, point, base_point])
        sym2 = 2 * point - id_n
        sym1 = 2 * base_point - id_n
        rot = self._general_linear.compose(sym2, sym1)
        return Matrices.bracket(self._general_linear.log(rot) / 2, base_point)

    def parallel_transport(
        self, tangent_vec, base_point, direction=None, end_point=None
    ):
        r"""Compute the parallel transport of a tangent vector.

        Closed-form solution for the parallel transport of a tangent vector
        along the geodesic between two points `base_point` and `end_point`
        or alternatively defined by :math:`t \mapsto exp_{(base\_point)}(
        t*direction)`.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., n, n]
            Tangent vector at base point to be transported.
        base_point : array-like, shape=[..., n, n]
            Point on the Grassmann manifold. Point to transport from.
        direction : array-like, shape=[..., n, n]
            Tangent vector at base point, along which the parallel transport
            is computed.
            Optional, default: None
        end_point : array-like, shape=[..., n, n]
            Point on the Grassmann manifold to transport to. Unused if
            `direction` is given.
            Optional, default: None

        Returns
        -------
        transported_tangent_vec: array-like, shape=[..., n, n]
            Transported tangent vector at `exp_(base_point)(direction)`.

        References
        ----------
        .. [BZA20]  Bendokat, Thomas, Ralf Zimmermann, and P.-A. Absil.
            “A Grassmann Manifold Handbook: Basic Geometry and Computational
            Aspects.” ArXiv:2011.13699 [Cs, Math], November 27, 2020.
            https://arxiv.org/abs/2011.13699.
        """
        if direction is None:
            if end_point is not None:
                direction = self.log(end_point, base_point)
            else:
                raise ValueError(
                    "Either an end_point or a direction must be given to define the"
                    " geodesic along which to transport."
                )
        expm = gs.linalg.expm
        mul = Matrices.mul
        rot = -Matrices.bracket(base_point, direction)
        return mul(expm(rot), tangent_vec, expm(-rot))

    def squared_dist(self, point_a, point_b):
        """Squared geodesic distance between two points.

        Parameters
        ----------
        point_a : array-like, shape=[..., dim]
            Point.
        point_b : array-like, shape=[..., dim]
            Point.

        Returns
        -------
        sq_dist : array-like, shape=[...,]
            Squared distance.
        """
        sdist_func = super().squared_dist

        def _squared_dist_grad_point_a(point_a, point_b):
            """Compute gradient of squared_dist wrt point_a."""
            return -2 * self.log(point_b, point_a)

        def _squared_dist_grad_point_b(point_a, point_b):
            """Compute gradient of squared_dist wrt point_b."""
            return -2 * self.log(point_a, point_b)

        @gs.autodiff.custom_gradient(
            _squared_dist_grad_point_a, _squared_dist_grad_point_b
        )
        def _squared_dist(point_a, point_b):
            """Compute geodesic distance between two points.

            Parameters
            ----------
            point_a : array-like, shape=[..., dim]
                Point.
            point_b : array-like, shape=[..., dim]
                Point.

            Returns
            -------
            _ : array-like, shape=[...,]
                Geodesic distance between point_a and point_b.
            """
            return sdist_func(point_a, point_b)

        return _squared_dist(point_a, point_b)

    def injectivity_radius(self, base_point=None):
        """Compute the radius of the injectivity domain.

        This is is the supremum of radii r for which the exponential map is a
        diffeomorphism from the open ball of radius r centered at the base
        point onto its image.
        In this case it is Pi / 2 everywhere.

        Parameters
        ----------
        base_point : array-like, shape=[..., n, n]
            Point on the manifold.

        Returns
        -------
        radius : array-like, shape=[...,]
            Injectivity radius.

        References
        ----------
        .. [BZA20]  Bendokat, Thomas, Ralf Zimmermann, and P.-A. Absil.
            “A Grassmann Manifold Handbook: Basic Geometry and
            Computational Aspects.”
            ArXiv:2011.13699 [Cs, Math], November 27, 2020.
            https://arxiv.org/abs/2011.13699.
        """
        radius = gs.array(gs.pi / 2)
        return repeat_out(self._space.point_ndim, radius, base_point)

    def inner_product(self, tangent_vec_a, tangent_vec_b, base_point=None):
        """Compute Frobenius inner-product of two tangent vectors.

        Coincides with the Frobenius metric.

        Parameters
        ----------
        tangent_vec_a : array-like, shape=[..., m, n]
            Tangent vector.
        tangent_vec_b : array-like, shape=[..., m, n]
            Tangent vector.
        base_point : array-like, shape=[..., m, n]
            Base point.
            Optional, default: None.

        Returns
        -------
        inner_prod : array-like, shape=[...,]
            Frobenius inner-product of tangent_vec_a and tangent_vec_b.
        """
        inner_prod = Matrices.frobenius_product(tangent_vec_a, tangent_vec_b)
        return repeat_out(
            self._space.point_ndim, inner_prod, tangent_vec_a, tangent_vec_b, base_point
        )

    def squared_norm(self, vector, base_point=None):
        """Compute the square of the norm of a vector.

        Squared norm of a vector associated to the inner product
        at the tangent space at a base point.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        sq_norm : array-like, shape=[...,]
            Squared norm.
        """
        sq_norm = gs.linalg.norm(vector, axis=(-2, -1)) ** 2
        return repeat_out(self._space.point_ndim, sq_norm, vector, base_point)

    def norm(self, vector, base_point=None):
        """Compute norm of a matrix.

        Norm of a matrix associated to the Frobenius inner product.

        Parameters
        ----------
        vector : array-like, shape=[..., dim]
            Vector.
        base_point : array-like, shape=[..., dim]
            Base point.
            Optional, default: None.

        Returns
        -------
        norm : array-like, shape=[...,]
            Norm.
        """
        norm = gs.linalg.norm(vector, axis=(-2, -1))
        return repeat_out(self._space.point_ndim, norm, vector, base_point)



"""Statistical Manifold of Poisson distributions with the Fisher metric.

Lead author: Tra My Nguyen.
"""

from scipy.special import factorial
from scipy.stats import poisson

import geomstats.backend as gs
from geomstats.geometry.base import VectorSpaceOpenSet
from geomstats.geometry.euclidean import Euclidean
from geomstats.geometry.riemannian_metric import RiemannianMetric
from geomstats.information_geometry.base import (
    InformationManifoldMixin,
    ScipyUnivariateRandomVariable,
)


class PoissonDistributions(InformationManifoldMixin, VectorSpaceOpenSet):
    """Class for the manifold of Poisson distributions.

    This is the parameter space of Poisson distributions
    i.e. [the half-line of positive reals].
    """

    def __init__(self, equip=True):
        super().__init__(
            dim=1,
            embedding_space=Euclidean(dim=1, equip=False),
            support_shape=(),
            equip=equip,
        )
        self._scp_rv = PoissonDistributionsRandomVariable(self)

    @staticmethod
    def default_metric():
        """Metric to equip the space with if equip is True."""
        return PoissonMetric

    def belongs(self, point, atol=gs.atol):
        """Evaluate if a point belongs to the manifold of Poisson distributions.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point to be checked.
        atol : float
            Tolerance to evaluate positivity.
            Optional, default: gs.atol

        Returns
        -------
        belongs : array-like, shape=[...,]
            Boolean indicating whether point represents an Poisson
            distribution.
        """
        belongs_shape = self.shape == point.shape[-self.point_ndim :]
        if not belongs_shape:
            shape = point.shape[: -self.point_ndim]
            return gs.zeros(shape, dtype=bool)
        return gs.squeeze(point >= -atol)

    def random_point(self, n_samples=1, bound=1.0):
        """Sample parameters of Poisson distributions.

        The uniform distribution on (0, bound) is used.

        Parameters
        ----------
        n_samples : int
            Number of samples.
            Optional, default: 1.
        bound : float
            Right-end ot the segment where Poisson parameters are sampled.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[n_samples,]
            Sample of points representing Poisson distributions.
        """
        size = (n_samples, self.dim) if n_samples != 1 else (self.dim,)
        return bound * gs.random.rand(*size)

    def projection(self, point, atol=gs.atol):
        """Project a point in ambient space to the open set.

        Return a point belonging to the half-line of positive reals
        within the given tolerance.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point in ambient space.
        atol : float
            Tolerance to evaluate positivity.

        Returns
        -------
        projected : array-like, shape=[..., 1]
            Projected point.
        """
        return gs.where(point < atol, atol, point)

    def sample(self, point, n_samples=1):
        """Sample from the Poisson distribution.

        Sample from the Poisson distribution with parameter provided
        by point.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point representing an Poisson distribution.
        n_samples : int
            Number of points to sample with each parameter in point.
            Optional, default: 1.

        Returns
        -------
        samples : array-like, shape=[..., n_samples]
            Sample from Poisson distributions.
        """
        return self._scp_rv.rvs(point, n_samples)

    def point_to_pdf(self, point):
        """Compute pmf associated to point.

        Compute the probability mass function of the Poisson
        distribution with parameters provided by point.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            Point representing an Poisson distribution (scale).

        Returns
        -------
        pdf : function
            Probability mass function of the Poisson distribution with
            scale parameter provided by point.
        """

        def pmf(k):
            """Generate parameterized function for Poisson pmf.

            Compute the probability mass function of the Poisson
            distribution with parameters provided by point.

            Parameters
            ----------
            k : array-like, shape=[n_samples,]
                Point in the sample space, i.e. an integer.

            Returns
            -------
            pmf_at_k : array-like, shape=[..., n_samples]
                Probability mass function of the Poisson distribution with
                parameters provided by point.
            """
            k = gs.reshape(gs.array(k), (-1,))
            return point**k * gs.exp(-point) / gs.from_numpy(factorial(k))

        return pmf


class PoissonMetric(RiemannianMetric):
    """Class for the Fisher information metric on Poisson distributions.

    References
    ----------
    .. [AM1981] Atkinson, C., & Mitchell, A. F. (1981). Rao's distance measure.
        Sankhyā: The Indian Journal of Statistics, Series A, 345-365.
    """

    def squared_dist(self, point_a, point_b):
        """Compute squared distance associated with the Poisson metric.

        Parameters
        ----------
        point_a : array-like, shape=[..., 1]
            Point representing an Poisson distribution (lambda parameter).
        point_b : array-like, shape=[..., 1]
            Point representing a Poisson distribution (lambda parameter).

        Returns
        -------
        squared_dist : array-like, shape=[...,]
            Squared distance between points point_a and point_b.
        """
        return gs.squeeze(4 * (point_a - 2 * gs.sqrt(point_a * point_b) + point_b))

    def metric_matrix(self, base_point):
        """Compute the metric matrix at the tangent space at base_point.

        Parameters
        ----------
        base_point : array-like, shape=[..., 1]
            Point representing a Poisson distribution.

        Returns
        -------
        mat : array-like, shape=[..., 1, 1]
            Metric matrix.
        """
        return gs.expand_dims(1 / base_point, axis=-1)

    @staticmethod
    def _geodesic_path(t, constant_a, constant_b):
        """Generate parameterized function for geodesic curve.

        Parameters
        ----------
        t : array-like, shape=[n_times,]
            Times at which to compute points of the geodesics.

        Returns
        -------
        geodesic : array-like, shape=[..., n_times, 1]
            Values of the geodesic at times t.
        """
        return gs.expand_dims((constant_a * t + constant_b) ** 2, axis=-1)

    def _geodesic_ivp(self, initial_point, initial_tangent_vec):
        """Solve geodesic initial value problem.

        Compute the parameterized function for the geodesic starting at
        initial_point with initial velocity given by initial_tangent_vec.

        Parameters
        ----------
        initial_point : array-like, shape=[..., 1]
            Initial point.

        initial_tangent_vec : array-like, shape=[..., 1]
            Tangent vector at initial point.

        Returns
        -------
        path : function
            Parameterized function for the geodesic curve starting at
            initial_point with velocity initial_tangent_vec.
        """
        constant_a = initial_tangent_vec / (2 * gs.sqrt(initial_point))
        constant_b = gs.sqrt(initial_point)
        return lambda t: self._geodesic_path(t, constant_a, constant_b)

    def _geodesic_bvp(self, initial_point, end_point):
        """Solve geodesic boundary problem.

        Compute the parameterized function for the geodesic starting at
        initial_point and ending at end_point.

        Parameters
        ----------
        initial_point : array-like, shape=[..., 1]
            Initial point.
        end_point : array-like, shape=[..., 1]
            End point.

        Returns
        -------
        path : function
            Parameterized function for the geodesic curve starting at
            initial_point and ending at end_point.
        """
        constant_a = gs.sqrt(end_point) - gs.sqrt(initial_point)
        constant_b = gs.sqrt(initial_point)
        return lambda t: self._geodesic_path(t, constant_a, constant_b)

    def exp(self, tangent_vec, base_point):
        """Compute exp map of a base point in tangent vector direction.

        Parameters
        ----------
        tangent_vec : array-like, shape=[..., 1]
            Tangent vector at base point.
        base_point : array-like, shape=[..., 1]
            Base point.

        Returns
        -------
        exp : array-like, shape=[..., 1]
            End point of the geodesic starting at base_point with
            initial velocity tangent_vec.
        """
        return (tangent_vec / (2 * gs.sqrt(base_point)) + gs.sqrt(base_point)) ** 2

    def log(self, point, base_point):
        """Compute log map using a base point and an end point.

        Parameters
        ----------
        point : array-like, shape=[..., 1]
            End point.
        base_point : array-like, shape=[..., 1]
            Base point.

        Returns
        -------
        tangent_vec : array-like, shape=[..., 1]
            Initial velocity of the geodesic starting at base_point and
            reaching point at time 1.
        """
        return 2 * gs.sqrt(base_point) * (gs.sqrt(point) - gs.sqrt(base_point))


class PoissonDistributionsRandomVariable(ScipyUnivariateRandomVariable):
    """A Poisson random variable."""

    def __init__(self, space):
        super().__init__(space, poisson.rvs)

    @staticmethod
    def _flatten_params(point, pre_flat_shape):
        flat_point = gs.reshape(gs.broadcast_to(point, pre_flat_shape), (-1,))
        return {"mu": flat_point}

    def pdf(self, x, point):
        """Evaluate the probability density function at x.

        Parameters
        ----------
        x : array-like, shape=[n_samples, *support_shape]
            Sample points at which to compute the probability density function.
        point : array-like, shape=[..., *space.shape]
            Point representing a distribution.

        Returns
        -------
        pdf_at_x : array-like, shape=[..., n_samples, *support_shape]
            Values of pdf at x for each value of the parameters provided
            by point.
        """
        return gs.from_numpy(
            poisson.pmf(x, point),
        )



import pytest

import geomstats.backend as gs
from geomstats.test.random import RandomDataGenerator, get_random_times
from geomstats.test.test_case import TestCase
from geomstats.test.vectorization import generate_vectorization_data


class _SolverTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)


class ExpSolverAgainstMetricTestCase(_SolverTestCase):
    """Exp solver against equipped space test case."""

    @pytest.mark.random
    def test_exp(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        res_ = self.exp_solver.exp(tangent_vec, base_point)
        res = self.space.metric.exp(tangent_vec, base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_geodesic_ivp(self, n_points, n_times, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)
        time = get_random_times(n_times)

        res_ = self.exp_solver.geodesic_ivp(tangent_vec, base_point)(time)
        res = self.space.metric.geodesic(base_point, initial_tangent_vec=tangent_vec)(
            time
        )
        self.assertAllClose(res, res_, atol=atol)


class ExpSolverComparisonTestCase(_SolverTestCase):
    """Exp solver against exp solver test case."""

    @pytest.mark.random
    def test_exp(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        res = self.exp_solver.exp(tangent_vec, base_point)
        res_ = self.cmp_exp_solver.exp(tangent_vec, base_point)

        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_geodesic_ivp(self, n_points, n_times, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)
        time = get_random_times(n_times)

        res = self.exp_solver.geodesic_ivp(tangent_vec, base_point)(time)
        res_ = self.cmp_exp_solver.geodesic_ivp(tangent_vec, base_point)(time)

        self.assertAllClose(res, res_, atol=atol)


class ExpSolverTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)

    def test_exp(self, tangent_vec, base_point, expected, atol):
        res = self.exp_solver.exp(tangent_vec, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_exp_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        expected = self.exp_solver.exp(tangent_vec, base_point)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point"],
            expected_name="expected",
            n_reps=n_reps,
            vectorization_type="repeat-0",
        )
        self._test_vectorization(vec_data)

    def test_geodesic_ivp(self, tangent_vec, base_point, time, expected, atol):
        res = self.exp_solver.geodesic_ivp(tangent_vec, base_point)(time)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_geodesic_ivp_vec(self, n_reps, n_times, atol):
        base_point = self.data_generator.random_point()
        tangent_vec = self.data_generator.random_tangent_vec(base_point)
        time = get_random_times(n_times)

        expected = self.exp_solver.geodesic_ivp(tangent_vec, base_point)(time)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    time=time,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point"],
            expected_name="expected",
            n_reps=n_reps,
            vectorization_type="repeat-0",
        )
        self._test_vectorization(vec_data)


class LogSolverAgainstMetricTestCase(_SolverTestCase):
    """Log solver against equipped space test case."""

    @pytest.mark.random
    def test_log(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        res_ = self.log_solver.log(end_point, base_point)
        res = self.space.metric.log(end_point, base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_geodesic_bvp(self, n_points, n_times, atol):
        base_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)
        time = get_random_times(n_times)

        res_ = self.log_solver.geodesic_bvp(end_point, base_point)(time)
        res = self.space.metric.geodesic(base_point, end_point=end_point)(time)
        self.assertAllClose(res, res_, atol=atol)


class LogSolverComparisonTestCase(_SolverTestCase):
    """Log solver against log solver test case."""

    @pytest.mark.random
    def test_log(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        res = self.log_solver.log(end_point, base_point)
        res_ = self.cmp_log_solver.log(end_point, base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_geodesic_bvp(self, n_points, n_times, atol):
        base_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)
        time = get_random_times(n_times)

        res = self.log_solver.geodesic_bvp(end_point, base_point)(time)
        res_ = self.cmp_log_solver.geodesic_bvp(end_point, base_point)(time)
        self.assertAllClose(res, res_, atol=atol)


class LogSolverTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)

    def test_log(self, point, base_point, expected, atol):
        res = self.log_solver.log(point, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_log_known_tangent_vec(self, n_points, atol):
        # NB: assumes proper implementation of exp
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        end_point = self.space.metric.exp(tangent_vec, base_point)

        self.test_log(end_point, base_point, tangent_vec, atol)

    def test_geodesic_bvp(self, point, base_point, time, expected, atol):
        res = self.log_solver.geodesic_bvp(point, base_point)(time)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_geodesic_bvp_vec(self, n_reps, n_times, atol):
        base_point = self.data_generator.random_point()
        point = self.data_generator.random_point()
        time = get_random_times(n_times)

        expected = self.log_solver.geodesic_bvp(point, base_point)(time)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    point=point,
                    base_point=base_point,
                    time=time,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["point", "base_point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_geodesic_bvp_known_geod(self, n_points, n_times, atol):
        # NB: assumes proper implementation of geodesic_ivp
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)
        time = get_random_times(n_times)

        point = self.space.metric.exp(tangent_vec, base_point)

        geod_point = self.space.metric.geodesic(
            base_point, initial_tangent_vec=tangent_vec
        )(time)

        self.test_geodesic_bvp(point, base_point, time, geod_point, atol)


class ExpSolverTypeCheck(_SolverTestCase):
    @pytest.mark.type
    def test_exp_type(self, n_points):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        res = self.exp_solver.exp(tangent_vec, base_point)
        self.assertTrue(gs.is_array(res), f"Wrong type: {type(res)}")

    @pytest.mark.type
    def test_geodesic_ivp_type(self, n_points, n_times):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)
        time = get_random_times(n_times)

        res = self.exp_solver.geodesic_ivp(tangent_vec, base_point)(time)
        self.assertTrue(gs.is_array(res), f"Wrong type: {type(res)}")


class LogSolverTypeCheckTestCase(_SolverTestCase):
    @pytest.mark.type
    def test_log_type(self, n_points):
        base_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        res = self.log_solver.log(end_point, base_point)
        self.assertTrue(gs.is_array(res), f"Wrong type: {type(res)}")

    @pytest.mark.type
    def test_geodesic_bvp_type(self, n_points, n_times):
        base_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)
        time = get_random_times(n_times)

        res = self.log_solver.geodesic_bvp(end_point, base_point)(time)
        self.assertTrue(gs.is_array(res), f"Wrong type: {type(res)}")



import math

import pytest

import geomstats.backend as gs
from geomstats.test.random import RandomDataGenerator, get_random_times
from geomstats.test.test_case import TestCase
from geomstats.test.utils import IdentityPointTransformer, PointTransformerFromDiffeo
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.mixins import GeodesicBVPTestCaseMixins
from geomstats.vectorization import get_batch_shape


class ConnectionTestCase(GeodesicBVPTestCaseMixins, TestCase):
    tangent_to_multiple = False
    is_metric = True

    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)

    def test_christoffels(self, base_point, expected, atol):
        res = self.space.metric.christoffels(base_point)
        self.assertAllClose(res, expected, atol=atol)

    def test_geodesic_equation(self, state, expected, atol):
        res = self.space.metric.geodesic_equation(state)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_geodesic_equation_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        state = gs.stack([base_point, tangent_vec])
        expected = self.space.metric.geodesic_equation(state)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    state=state,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["state"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_exp(self, tangent_vec, base_point, expected, atol):
        res = self.space.metric.exp(tangent_vec, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_exp_belongs(self, n_points, atol):
        """Check exponential gives point in the manifold.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        point = self.space.metric.exp(tangent_vec, base_point)

        res = self.space.belongs(point, atol=atol)
        expected_shape = get_batch_shape(self.space.point_ndim, base_point)
        expected = gs.ones(expected_shape, dtype=bool)
        self.assertAllEqual(res, expected)

    def test_log(self, point, base_point, expected, atol):
        res = self.space.metric.log(point, base_point)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.random
    def test_log_is_tangent(self, n_points, atol):
        """Check logarithm gives a tangent vector.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        base_point = self.data_generator.random_point(n_points)
        point = self.data_generator.random_point(n_points)

        tangent_vec = self.space.metric.log(point, base_point)

        res = self.space.is_tangent(tangent_vec, base_point, atol=atol)
        expected_shape = get_batch_shape(self.space.point_ndim, base_point)
        expected = gs.ones(expected_shape, dtype=bool)
        self.assertAllEqual(res, expected)

    @pytest.mark.random
    def test_exp_after_log(self, n_points, atol):
        """Check exp and log are inverse.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        base_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)

        tangent_vec = self.space.metric.log(end_point, base_point)
        end_point_ = self.space.metric.exp(tangent_vec, base_point)

        self.assertAllClose(end_point_, end_point, atol=atol)

    @pytest.mark.random
    def test_log_after_exp(self, n_points, atol):
        """Check log and exp are inverse.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        end_point = self.space.metric.exp(tangent_vec, base_point)
        tangent_vec_ = self.space.metric.log(end_point, base_point)

        self.assertAllClose(tangent_vec_, tangent_vec, atol=atol)

    def test_riemann_tensor(self, base_point, expected, atol):
        res = self.space.metric.riemann_tensor(base_point)
        self.assertAllClose(res, expected, atol=atol)

    def test_curvature(
        self, tangent_vec_a, tangent_vec_b, tangent_vec_c, base_point, expected, atol
    ):
        res = self.space.metric.curvature(
            tangent_vec_a, tangent_vec_b, tangent_vec_c, base_point
        )
        self.assertAllClose(res, expected, atol=atol)

    def test_ricci_tensor(self, base_point, expected, atol):
        res = self.space.metric.ricci_tensor(base_point)
        self.assertAllClose(res, expected, atol=atol)

    def test_directional_curvature(
        self, tangent_vec_a, tangent_vec_b, base_point, expected, atol
    ):
        res = self.space.metric.directional_curvature(
            tangent_vec_a, tangent_vec_b, base_point
        )
        self.assertAllClose(res, expected, atol=atol)

    def test_curvature_derivative(
        self,
        tangent_vec_a,
        tangent_vec_b,
        tangent_vec_c,
        tangent_vec_d,
        base_point,
        expected,
        atol,
    ):
        res = self.space.metric.curvature_derivative(
            tangent_vec_a, tangent_vec_b, tangent_vec_c, tangent_vec_d, base_point
        )
        self.assertAllClose(res, expected, atol=atol)

    def test_directional_curvature_derivative(
        self, tangent_vec_a, tangent_vec_b, base_point, expected, atol
    ):
        res = self.space.metric.directional_curvature_derivative(
            tangent_vec_a, tangent_vec_b, base_point
        )
        self.assertAllClose(res, expected, atol=atol)

    def test_geodesic(
        self,
        initial_point,
        time,
        expected,
        atol,
        end_point=None,
        initial_tangent_vec=None,
    ):
        geod_func = self.space.metric.geodesic(
            initial_point, end_point=end_point, initial_tangent_vec=initial_tangent_vec
        )
        res = geod_func(time)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_geodesic_ivp_vec(self, n_reps, n_times, atol):
        initial_point = self.data_generator.random_point()
        initial_tangent_vec = self.data_generator.random_tangent_vec(initial_point)
        time = get_random_times(n_times)

        expected = self.space.metric.geodesic(
            initial_point, initial_tangent_vec=initial_tangent_vec
        )(time)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    initial_point=initial_point,
                    initial_tangent_vec=initial_tangent_vec,
                    time=time,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["initial_point", "initial_tangent_vec"],
            expected_name="expected",
            n_reps=n_reps,
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-1",
        )
        self._test_vectorization(vec_data, test_fnc_name="test_geodesic")

    @pytest.mark.random
    def test_geodesic_ivp_belongs(self, n_points, n_times, atol):
        """Check geodesic belongs to manifold.

        This is for geodesics defined by the initial value problem (ivp).

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        initial_point = self.data_generator.random_point(n_points)
        initial_tangent_vec = self.data_generator.random_tangent_vec(initial_point)

        time = get_random_times(n_times)

        geod_func = self.space.metric.geodesic(
            initial_point, initial_tangent_vec=initial_tangent_vec
        )

        points = geod_func(time)

        res = self.space.belongs(gs.reshape(points, (-1, *self.space.shape)), atol=atol)

        expected_shape = (
            math.prod(get_batch_shape(self.space.point_ndim, initial_point)) * n_times,
        )

        expected = gs.ones(expected_shape, dtype=bool)
        self.assertAllEqual(res, expected)

    @pytest.mark.random
    def test_exp_geodesic_ivp(self, n_points, atol):
        """Check end point of a geodesic matches exponential.

        Parameters
        ----------
        n_points : int
            Number of random points to generate.
        atol : float
            Absolute tolerance.
        """
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        geod_func = self.space.metric.geodesic(
            base_point, initial_tangent_vec=tangent_vec
        )

        end_point = self.space.metric.exp(tangent_vec, base_point)
        end_point_ = gs.squeeze(geod_func(1.0), axis=-(self.space.point_ndim + 1))

        self.assertAllClose(end_point_, end_point, atol=atol)

    def test_parallel_transport(
        self, tangent_vec, base_point, expected, atol, direction=None, end_point=None
    ):
        res = self.space.metric.parallel_transport(
            tangent_vec,
            base_point,
            direction=direction,
            end_point=end_point,
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_parallel_transport_ivp_vec(self, n_reps, atol):
        base_point = self.data_generator.random_point()
        tangent_vec = self.data_generator.random_tangent_vec(base_point)
        direction = self.data_generator.random_tangent_vec(base_point)

        expected = self.space.metric.parallel_transport(
            tangent_vec, base_point, direction=direction
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    direction=direction,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point", "direction"],
            expected_name="expected",
            n_reps=n_reps,
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-0-2",
        )
        self._test_vectorization(vec_data, test_fnc_name="test_parallel_transport")

    @pytest.mark.vec
    def test_parallel_transport_bvp_vec(self, n_reps, atol):
        base_point, end_point = self.data_generator.random_point(2)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        expected = self.space.metric.parallel_transport(
            tangent_vec, base_point, end_point=end_point
        )

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    tangent_vec=tangent_vec,
                    base_point=base_point,
                    end_point=end_point,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["tangent_vec", "base_point", "end_point"],
            expected_name="expected",
            n_reps=n_reps,
            vectorization_type="sym" if self.tangent_to_multiple else "repeat-0",
        )
        self._test_vectorization(vec_data, test_fnc_name="test_parallel_transport")

    @pytest.mark.random
    def test_parallel_transport_bvp_transported_is_tangent(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        transported = self.space.metric.parallel_transport(
            tangent_vec, base_point, end_point=end_point
        )

        res = self.space.is_tangent(transported, end_point, atol=atol)

        expected_shape = get_batch_shape(self.space.point_ndim, base_point)
        expected = gs.ones(expected_shape, dtype=bool)

        self.assertAllEqual(res, expected)

    @pytest.mark.random
    def test_parallel_transport_ivp_transported_is_tangent(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        direction = self.data_generator.random_tangent_vec(base_point)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        transported = self.space.metric.parallel_transport(
            tangent_vec, base_point, direction=direction
        )

        end_point = self.space.metric.exp(direction, base_point)

        res = self.space.is_tangent(transported, end_point, atol=atol)

        expected_shape = get_batch_shape(self.space.point_ndim, base_point)
        expected = gs.ones(expected_shape, dtype=bool)

        self.assertAllEqual(res, expected)

    def test_injectivity_radius(self, base_point, expected, atol):
        res = self.space.metric.injectivity_radius(base_point)
        self.assertAllClose(res, expected, atol=atol)


class ConnectionComparisonTestCase(TestCase):
    def setup_method(self):
        if not hasattr(self, "data_generator"):
            self.data_generator = RandomDataGenerator(self.space)

        if not hasattr(self, "point_transformer") and hasattr(self, "diffeo"):
            self.point_transformer = PointTransformerFromDiffeo(self.diffeo)

        if not hasattr(self, "point_transformer"):
            self.point_transformer = IdentityPointTransformer()

    def test_christoffels(self, base_point, atol):
        res = self.space.metric.christoffels(base_point)
        res_ = self.other_space.metric.christoffels(base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_christoffels_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        self.test_christoffels(base_point, atol)

    def test_exp(self, tangent_vec, base_point, atol):
        base_point_ = self.point_transformer.transform_point(base_point)
        tangent_vec_ = self.point_transformer.transform_tangent_vec(
            tangent_vec, base_point
        )

        res = self.space.metric.exp(tangent_vec, base_point)
        res_ = self.other_space.metric.exp(tangent_vec_, base_point_)

        res_ = self.point_transformer.inverse_transform_point(res_)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_exp_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        self.test_exp(tangent_vec, base_point, atol)

    def test_log(self, point, base_point, atol):
        base_point_ = self.point_transformer.transform_point(base_point)
        point_ = self.point_transformer.transform_point(point)

        res = self.space.metric.log(point, base_point)
        res_ = self.other_space.metric.log(point_, base_point_)

        res_ = self.point_transformer.inverse_transform_tangent_vec(res_, base_point_)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_log_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        point = self.data_generator.random_point(n_points)

        self.test_log(point, base_point, atol)

    def test_riemann_tensor(self, base_point, atol):
        base_point_ = self.point_transformer.transform_point(base_point)

        res = self.space.metric.riemann_tensor(base_point)
        res_ = self.other_space.metric.riemann_tensor(base_point_)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_riemann_tensor_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        self.test_riemann_tensor(base_point, atol)

    def test_curvature(
        self, tangent_vec_a, tangent_vec_b, tangent_vec_c, base_point, atol
    ):
        base_point_ = self.point_transformer.transform_point(base_point)
        tangent_vec_a_ = self.point_transformer.transform_tangent_vec(
            tangent_vec_a, base_point
        )
        tangent_vec_b_ = self.point_transformer.transform_tangent_vec(
            tangent_vec_b, base_point
        )
        tangent_vec_c_ = self.point_transformer.transform_tangent_vec(
            tangent_vec_c, base_point
        )

        res = self.space.metric.curvature(
            tangent_vec_a, tangent_vec_b, tangent_vec_c, base_point
        )
        res_ = self.other_space.metric.curvature(
            tangent_vec_a_, tangent_vec_b_, tangent_vec_c_, base_point_
        )
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_curvature_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_c = self.data_generator.random_tangent_vec(base_point)

        self.test_curvature(
            tangent_vec_a, tangent_vec_b, tangent_vec_c, base_point, atol
        )

    def test_ricci_tensor(self, base_point, atol):
        res = self.space.metric.ricci_tensor(base_point)
        res_ = self.other_space.metric.ricci_tensor(base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_ricci_tensor_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        self.test_ricci_tensor(base_point, atol)

    def test_directional_curvature(
        self, tangent_vec_a, tangent_vec_b, base_point, atol
    ):
        res = self.space.metric.directional_curvature(
            tangent_vec_a, tangent_vec_b, base_point
        )
        res_ = self.other_space.metric.directional_curvature(
            tangent_vec_a, tangent_vec_b, base_point
        )
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_directional_curvature_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        self.test_directional_curvature(tangent_vec_a, tangent_vec_b, base_point, atol)

    def test_curvature_derivative(
        self,
        tangent_vec_a,
        tangent_vec_b,
        tangent_vec_c,
        tangent_vec_d,
        base_point,
        atol,
    ):
        res = self.space.metric.curvature_derivative(
            tangent_vec_a, tangent_vec_b, tangent_vec_c, tangent_vec_d, base_point
        )
        res_ = self.other_space.metric.curvature_derivative(
            tangent_vec_a, tangent_vec_b, tangent_vec_c, tangent_vec_d, base_point
        )
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_curvature_derivative_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_c = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_d = self.data_generator.random_tangent_vec(base_point)

        self.test_curvature_derivative(
            tangent_vec_a, tangent_vec_b, tangent_vec_c, tangent_vec_d, base_point, atol
        )

    def test_directional_curvature_derivative(
        self, tangent_vec_a, tangent_vec_b, base_point, atol
    ):
        res = self.space.metric.directional_curvature_derivative(
            tangent_vec_a, tangent_vec_b, base_point
        )
        res_ = self.other_space.metric.directional_curvature_derivative(
            tangent_vec_a, tangent_vec_b, base_point
        )
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_directional_curvature_derivative_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(base_point)
        tangent_vec_b = self.data_generator.random_tangent_vec(base_point)

        self.test_directional_curvature_derivative(
            tangent_vec_a, tangent_vec_b, base_point, atol
        )

    def test_geodesic_bvp(self, initial_point, end_point, time, atol):
        initial_point_ = self.point_transformer.transform_point(initial_point)
        end_point_ = self.point_transformer.transform_point(end_point)

        res = self.space.metric.geodesic(initial_point, end_point=end_point)(time)
        res_ = self.other_space.metric.geodesic(initial_point_, end_point=end_point_)(
            time
        )

        res_ = self.point_transformer.inverse_transform_point(res_)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_geodesic_bvp_random(self, n_points, n_times, atol):
        initial_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)
        time = get_random_times(n_times)

        self.test_geodesic_bvp(initial_point, end_point, time, atol)

    def test_geodesic_ivp(self, initial_point, initial_tangent_vec, time, atol):
        initial_point_ = self.point_transformer.transform_point(initial_point)
        initial_tangent_vec_ = self.point_transformer.transform_tangent_vec(
            initial_tangent_vec, initial_point
        )

        res = self.space.metric.geodesic(
            initial_point, initial_tangent_vec=initial_tangent_vec
        )(time)

        res_ = self.other_space.metric.geodesic(
            initial_point_, initial_tangent_vec=initial_tangent_vec_
        )(time)

        res_ = self.point_transformer.inverse_transform_point(res_)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_geodesic_ivp_random(self, n_points, n_times, atol):
        initial_point = self.data_generator.random_point(n_points)
        initial_tangent_vec = self.data_generator.random_tangent_vec(initial_point)
        time = get_random_times(n_times)

        self.test_geodesic_ivp(initial_point, initial_tangent_vec, time, atol)

    def test_parallel_transport_ivp(self, base_point, tangent_vec, direction, atol):
        base_point_ = self.point_transformer.transform_point(base_point)
        tangent_vec_ = self.point_transformer.transform_tangent_vec(
            tangent_vec, base_point
        )
        direction_ = self.point_transformer.transform_tangent_vec(direction, base_point)

        res = self.space.metric.parallel_transport(
            tangent_vec, base_point, direction=direction
        )
        res_ = self.other_space.metric.parallel_transport(
            tangent_vec_, base_point_, direction=direction_
        )

        end_point_ = self.other_space.metric.exp(direction_, base_point_)
        res_ = self.point_transformer.inverse_transform_tangent_vec(res_, end_point_)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_parallel_transport_ivp_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)
        direction = self.data_generator.random_tangent_vec(base_point)

        self.test_parallel_transport_ivp(base_point, tangent_vec, direction, atol)

    def test_parallel_transport_bvp(self, base_point, end_point, tangent_vec, atol):
        base_point_ = self.point_transformer.transform_point(base_point)
        end_point_ = self.point_transformer.transform_point(end_point)
        tangent_vec_ = self.point_transformer.transform_tangent_vec(
            tangent_vec, base_point
        )

        res = self.space.metric.parallel_transport(
            tangent_vec, base_point, end_point=end_point
        )
        res_ = self.other_space.metric.parallel_transport(
            tangent_vec_, base_point_, end_point=end_point_
        )

        res_ = self.point_transformer.inverse_transform_tangent_vec(res_, end_point_)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_parallel_transport_bvp_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        end_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        self.test_parallel_transport_bvp(base_point, end_point, tangent_vec, atol)

    def test_injectivity_radius(self, base_point, atol):
        res = self.space.metric.injectivity_radius(base_point)
        res_ = self.other_space.metric.injectivity_radius(base_point)
        self.assertAllClose(res, res_, atol=atol)

    @pytest.mark.random
    def test_injectivity_radius_random(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)

        self.test_injectivity_radius(base_point, atol)



import pytest

import geomstats.backend as gs
from geomstats.test.vectorization import generate_vectorization_data
from geomstats.test_cases.geometry.manifold import ManifoldTestCase
from geomstats.test_cases.geometry.mixins import GroupExpTestCaseMixins
from geomstats.vectorization import repeat_point


class _LieGroupTestCaseMixins(GroupExpTestCaseMixins):
    def test_compose(self, point_a, point_b, expected, atol):
        composed = self.space.compose(point_a, point_b)
        self.assertAllClose(composed, expected, atol=atol)

    def test_inverse(self, point, expected, atol):
        inverse = self.space.inverse(point)
        self.assertAllClose(inverse, expected, atol=atol)

    @pytest.mark.random
    def test_compose_with_inverse_is_identity(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        inverse = self.space.inverse(point)

        identity = self.space.identity
        if n_points > 1:
            identity = gs.broadcast_to(identity, (n_points, *identity.shape))

        identity_ = self.space.compose(point, inverse)
        self.assertAllClose(identity_, identity, atol=atol)

        identity_ = self.space.compose(inverse, point)
        self.assertAllClose(identity_, identity, atol=atol)

    @pytest.mark.random
    def test_compose_with_identity_is_itself(self, n_points, atol):
        point = self.data_generator.random_point(n_points)

        point_ = self.space.compose(point, self.space.identity)
        self.assertAllClose(point_, point, atol=atol)

        point_ = self.space.compose(self.space.identity, point)
        self.assertAllClose(point_, point, atol=atol)

    def test_log(self, point, base_point, expected, atol):
        vec = self.space.log(point, base_point)
        self.assertAllClose(vec, expected, atol=atol)

    @pytest.mark.random
    def test_exp_after_log(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        base_point = self.data_generator.random_point(n_points)

        vec = self.space.log(point, base_point)
        point_ = self.space.exp(vec, base_point)

        self.assertAllClose(point_, point, atol=atol)

    @pytest.mark.random
    def test_log_after_exp(self, n_points, atol):
        base_point = self.data_generator.random_point(n_points)
        tangent_vec = self.data_generator.random_tangent_vec(base_point)

        point = self.space.exp(tangent_vec, base_point)
        tangent_vec_ = self.space.log(point, base_point)

        self.assertAllClose(tangent_vec_, tangent_vec, atol=atol)

    @pytest.mark.random
    def test_to_tangent_at_identity_belongs_to_lie_algebra(self, n_points, atol):
        tangent_vec = self.data_generator.random_tangent_vec(
            repeat_point(self.space.identity, n_points)
        )

        res = self.space.lie_algebra.belongs(tangent_vec, atol=atol)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(res, expected)

    def test_tangent_translation_map(
        self, point, left, inverse, tangent_vec, expected, atol
    ):
        res = self.space.tangent_translation_map(point, left=left, inverse=inverse)(
            tangent_vec
        )
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_tangent_translation_map_vec(self, n_reps, left, inverse, atol):
        point = self.data_generator.random_point()
        tangent_vec = self.data_generator.random_tangent_vec(point)

        expected = self.space.tangent_translation_map(
            point, left=left, inverse=inverse
        )(tangent_vec)

        vec_data = generate_vectorization_data(
            data=[
                dict(
                    point=point,
                    tangent_vec=tangent_vec,
                    left=left,
                    inverse=inverse,
                    expected=expected,
                    atol=atol,
                )
            ],
            arg_names=["point", "tangent_vec"],
            expected_name="expected",
            vectorization_type="repeat-1",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_tangent_translation_map_is_tangent(self, n_points, left, atol):
        group_elem_1 = self.data_generator.random_point(n_points)
        group_elem_2 = self.data_generator.random_point(n_points)

        tangent_vec_11 = self.data_generator.random_tangent_vec(group_elem_1)

        dtrans_2 = self.space.tangent_translation_map(
            group_elem_2, left=left, inverse=False
        )

        if left:
            group_elem_3 = self.space.compose(group_elem_2, group_elem_1)
        else:
            group_elem_3 = self.space.compose(group_elem_1, group_elem_2)

        tangent_vec_31 = dtrans_2(tangent_vec_11)

        res = self.space.is_tangent(tangent_vec_31, group_elem_3, atol=atol)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(res, expected)

    def test_lie_bracket(
        self, tangent_vec_a, tangent_vec_b, base_point, expected, atol
    ):
        bracket = self.space.lie_bracket(tangent_vec_a, tangent_vec_b, base_point)
        self.assertAllClose(bracket, expected, atol=atol)

    @pytest.mark.random
    def test_lie_bracket_belongs(self, n_points, atol):
        identity = repeat_point(self.space.identity, n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(identity)
        tangent_vec_b = self.data_generator.random_tangent_vec(identity)

        tangent_vec_c = self.space.lie_bracket(tangent_vec_a, tangent_vec_b)

        res = self.space.lie_algebra.belongs(tangent_vec_c, atol=atol)
        expected = gs.ones(n_points, dtype=bool)
        self.assertAllEqual(res, expected)

    @pytest.mark.random
    def test_lie_bracket_antisymmetry(self, n_points, atol):
        identity = repeat_point(self.space.identity, n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(identity)
        tangent_vec_b = self.data_generator.random_tangent_vec(identity)

        tangent_vec_c = self.space.lie_bracket(tangent_vec_a, tangent_vec_b)
        tangent_vec_c_ = -self.space.lie_bracket(tangent_vec_b, tangent_vec_a)
        self.assertAllClose(tangent_vec_c, tangent_vec_c_, atol=atol)

    @pytest.mark.random
    def test_lie_bracket_jacobi_identity(self, n_points, atol):
        identity = repeat_point(self.space.identity, n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(identity)
        tangent_vec_b = self.data_generator.random_tangent_vec(identity)
        tangent_vec_c = self.data_generator.random_tangent_vec(identity)

        lhs = self.space.lie_bracket(
            self.space.lie_bracket(tangent_vec_a, tangent_vec_b), tangent_vec_c
        )
        rhs_1 = self.space.lie_bracket(
            tangent_vec_a, self.space.lie_bracket(tangent_vec_b, tangent_vec_c)
        )
        rhs_2 = -self.space.lie_bracket(
            tangent_vec_b, self.space.lie_bracket(tangent_vec_a, tangent_vec_c)
        )
        rhs = rhs_1 + rhs_2

        self.assertAllClose(lhs, rhs, atol=atol)

    @pytest.mark.random
    def test_lie_bracket_bilinearity(self, n_points, atol):
        identity = repeat_point(self.space.identity, n_points)
        tangent_vec_a = self.data_generator.random_tangent_vec(identity)
        tangent_vec_b = self.data_generator.random_tangent_vec(identity)
        tangent_vec_c = self.data_generator.random_tangent_vec(identity)

        a, b = gs.random.rand(2)

        lhs = self.space.lie_bracket(
            a * tangent_vec_a + b * tangent_vec_b, tangent_vec_c
        )
        rhs_1 = a * self.space.lie_bracket(tangent_vec_a, tangent_vec_c)
        rhs_2 = b * self.space.lie_bracket(tangent_vec_b, tangent_vec_c)
        rhs = rhs_1 + rhs_2
        self.assertAllClose(lhs, rhs, atol=atol)

        lhs = self.space.lie_bracket(
            tangent_vec_c, a * tangent_vec_a + b * tangent_vec_b
        )
        rhs_1 = a * self.space.lie_bracket(tangent_vec_c, tangent_vec_a)
        rhs_2 = b * self.space.lie_bracket(tangent_vec_c, tangent_vec_b)
        rhs = rhs_1 + rhs_2

        self.assertAllClose(lhs, rhs, atol=atol)

    def test_identity(self, expected, atol):
        self.assertAllClose(self.space.identity, expected, atol=atol)


class MatrixLieGroupTestCase(_LieGroupTestCaseMixins, ManifoldTestCase):
    pass


class LieGroupTestCase(_LieGroupTestCaseMixins, ManifoldTestCase):
    # TODO: exp and log not from identity: are they enough tested with log and exp?

    def test_jacobian_translation(self, point, left, expected, atol):
        res = self.space.jacobian_translation(point, left=left)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_jacobian_translation_vec(self, n_reps, left, atol):
        point = self.data_generator.random_point()
        expected = self.space.jacobian_translation(point, left=left)

        vec_data = generate_vectorization_data(
            data=[dict(point=point, left=left, expected=expected, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_exp_from_identity(self, tangent_vec, expected, atol):
        res = self.space.exp_from_identity(tangent_vec)
        self.assertAllClose(res, expected, atol=atol)

    @pytest.mark.vec
    def test_exp_from_identity_vec(self, n_reps, atol):
        tangent_vec = self.data_generator.random_tangent_vec(self.space.identity)
        expected = self.space.exp_from_identity(tangent_vec)

        vec_data = generate_vectorization_data(
            data=[dict(tangent_vec=tangent_vec, expected=expected, atol=atol)],
            arg_names=["tangent_vec"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    def test_log_from_identity(self, point, expected, atol):
        vec = self.space.log_from_identity(point)
        self.assertAllClose(vec, expected, atol=atol)

    @pytest.mark.vec
    def test_log_from_identity_vec(self, n_reps, atol):
        point = self.data_generator.random_point()
        expected = self.space.log_from_identity(point)

        vec_data = generate_vectorization_data(
            data=[dict(point=point, expected=expected, atol=atol)],
            arg_names=["point"],
            expected_name="expected",
            n_reps=n_reps,
        )
        self._test_vectorization(vec_data)

    @pytest.mark.random
    def test_exp_from_identity_after_log_from_identity(self, n_points, atol):
        point = self.data_generator.random_point(n_points)
        tangent_vec = self.space.log_from_identity(point)

        point_ = self.space.exp_from_identity(tangent_vec)
        self.assertAllClose(point_, point, atol=atol)

    @pytest.mark.random
    def test_log_from_identity_after_exp_from_identity(self, n_points, atol):
        tangent_vec = self.data_generator.random_tangent_vec(
            repeat_point(self.space.identity, n_points)
        )

        point = self.space.exp_from_identity(tangent_vec)
        tangent_vec_ = self.space.log_from_identity(point)

        self.assertAllClose(tangent_vec_, tangent_vec, atol=atol)
